module ddhmc_io_class
use lattice_class
use timer_class
implicit none
private
public :: ddhmc_node
public :: config_prof_ddhmc
public :: new,delete
public :: plaquette
public :: read_ddhmc_config
public :: convert_and_save_config

!integer, parameter :: COL=3,NDIM=4
!integer, parameter :: DP=KIND(1.0d0)

!integer, parameter :: isx(NDIM) = (/1,0,0,0/)
!integer, parameter :: isy(NDIM) = (/0,1,0,0/)
!integer, parameter :: isz(NDIM) = (/0,0,1,0/)
!integer, parameter :: ist(NDIM) = (/0,0,0,1/)

type su3mat
  complex(DP) :: u(COL,COL)
end type

type gsf_eo_deo_b
  type(su3mat), allocatable :: s(:,:,:,:)
end type

type gsf_eo_deo
  type(gsf_eo_deo_b), allocatable :: b(:)
end type

type gvf_eo_deo
  type(gsf_eo_deo) :: mu(NDIM)
end type

type gvf_deo
  type(gvf_eo_deo) :: eo(0:1)
end type

type gvf
  type(gvf_deo) :: deo(0:1)
end type

interface new
  module procedure new_gvf
end interface

interface delete
  module procedure delete_gvf
end interface

type config_prof_quark
  real(DP) :: K
  real(DP) :: Csw
  integer  :: nf
end type

type config_prof_ddhmc
  integer :: cNTX,cNTY,cNTZ,cNTT
  integer :: cNDIMX,cNDIMY,cNDIMZ,cNDIMT
  integer :: cnodeid
  integer :: cipsite(NDIM)
  integer :: rand_seed
  integer :: traj
  integer :: num_quark
  real(DP) :: plq
  real(DP) :: beta
  real(DP) :: cRG0, u0
  type(config_prof_quark), allocatable :: quark(:)
  character(256) :: cdate,ctime,czone,comment
end type

type node_params
  integer :: NTX
  integer :: NTY
  integer :: NTZ
  integer :: NTT
  integer :: NDIMX
  integer :: NDIMY
  integer :: NDIMZ
  integer :: NDIMT
  integer :: NDX
  integer :: NDY
  integer :: NDZ
  integer :: NDT
  integer :: NPU
  integer :: NDM
  integer :: NX
  integer :: NY
  integer :: NZ
  integer :: NT
  integer :: NBX
  integer :: NBY
  integer :: NBZ
  integer :: NBT
  integer :: NBTH
  integer :: ipeo
  integer :: ipdeo
  integer, allocatable :: idmod(:,:)
  integer, allocatable :: iddeo(:,:)
end type

type(node_params), save :: ddhmc_node

interface new
  module procedure new_node_params
end interface

contains

subroutine new_gvf(u)
  type(gvf), intent(inout) :: u
  integer :: ideo,ieo,mu,id
  do ideo=0,1
  do ieo=0,1
  do mu=1,NDIM
    allocate(u%deo(ideo)%eo(ieo)%mu(mu)%b(0:ddhmc_node%NDM-1))
    do id=0,ddhmc_node%NDM-1
      allocate(u%deo(ideo)%eo(ieo)%mu(mu)%b(id)%s(0:ddhmc_node%NBTH,ddhmc_node%NBZ,ddhmc_node%NBY,ddhmc_node%NBX))
    enddo
  enddo
  enddo
  enddo
  return
end subroutine

subroutine delete_gvf(u)
  type(gvf), intent(inout) :: u
  integer :: ideo,ieo,mu,id
  do ideo=0,1
  do ieo=0,1
  do mu=1,NDIM
    do id=0,ddhmc_node%NDM-1
      deallocate(u%deo(ideo)%eo(ieo)%mu(mu)%b(id)%s)
    enddo
    deallocate(u%deo(ideo)%eo(ieo)%mu(mu)%b)
  enddo
  enddo
  enddo
  return
end subroutine

subroutine new_node_params(this)
  type(node_params), intent(inout) :: this
  this%NPU = this%NDIMX*this%NDIMY*this%NDIMZ*this%NDIMT
  this%NDM = (this%NDX*this%NDY*this%NDZ*this%NDT)/2
  this%NX  = this%NTX/this%NDIMX
  this%NY  = this%NTY/this%NDIMY
  this%NZ  = this%NTZ/this%NDIMZ
  this%NT  = this%NTT/this%NDIMT
  this%NBX = this%NX/this%NDX
  this%NBY = this%NY/this%NDY
  this%NBZ = this%NZ/this%NDZ
  this%NBT = this%NT/this%NDT
  this%NBTH = this%NBT/2
  return
end subroutine

subroutine get_node_params(this,ipsite)
  type(node_params), intent(inout) :: this
  integer, intent(in) :: ipsite(NDIM)
  integer :: idx,idy,idz,idt,ideo,idd,id
  if (allocated(this%idmod)) deallocate(this%idmod)
  if (allocated(this%iddeo)) deallocate(this%iddeo)
  allocate(this%idmod(0:this%NDM-1,0:1))
  allocate(this%iddeo(0:this%NDM-1,0:1))
  this%ipeo  = mod(ipsite(1)*this%NX +ipsite(2)*this%NY +ipsite(3)*this%NZ +ipsite(4)*this%NT ,2)
  this%ipdeo = mod(ipsite(1)*this%NDX+ipsite(2)*this%NDY+ipsite(3)*this%NDZ+ipsite(4)*this%NDT,2)
  do idx=0,this%NDX-1
  do idy=0,this%NDY-1
  do idz=0,this%NDZ-1
  do idt=0,this%NDT-1
    ideo = mod(this%ipdeo+idt+idz+idy+idx,2)
    idd  = idt+idz*this%NDT+idy*this%NDT*this%NDZ+idx*this%NDT*this%NDZ*this%NDY
    id   = idd/2
    this%idmod(id,ideo) = mod(idd,2)
    this%iddeo(id,ideo) = mod(this%ipeo+idx*this%NBX+idy*this%NBY+idz*this%NBZ+idt*this%NBT,2)
  enddo
  enddo
  enddo
  enddo
  return
end subroutine

subroutine read_config_prof(this,iout)
  type(config_prof_ddhmc), intent(inout) :: this
  integer, intent(in) :: iout
  integer :: iqk

  if (allocated(this%quark)) deallocate(this%quark)

  read(iout) this%cNTX,this%cNTY,this%cNTZ,this%cNTT
  read(iout) this%cNDIMX,this%cNDIMY,this%cNDIMZ,this%cNDIMT
  read(iout) this%cnodeid
  read(iout) this%cipsite(1),this%cipsite(2),this%cipsite(3),this%cipsite(4)
  read(iout) this%plq
  read(iout) this%cdate
  read(iout) this%ctime
  read(iout) this%czone
  read(iout) this%comment
  read(iout) this%rand_seed
  read(iout) this%traj
  read(iout) this%beta
  read(iout) this%cRG0
  read(iout) this%u0
  read(iout) this%num_quark
  allocate(this%quark(this%num_quark))
  do iqk=1,this%num_quark
    read(iout) this%quark(iqk)%nf
    read(iout) this%quark(iqk)%K
    read(iout) this%quark(iqk)%Csw
  enddo

  write(*,'(80("-"))')
  write(*,'("config: DATE:",A)') TRIM(this%cdate)
  write(*,'("config: TIME:",A)') TRIM(this%ctime)
  write(*,'("config: ZONE:",A)') TRIM(this%czone)
  write(*,'("config: COMM:",A)') TRIM(this%comment)
  write(*,'("conifg:          TRAJ:",I6)')  this%traj
  write(*,'("config:  LATTICE XYZT:",4I3)') this%cNTX,this%cNTY,this%cNTZ,this%cNTT
  write(*,'("config:  NODEDIM XYZT:",4I3)') this%cNDIMX,this%cNDIMY,this%cNDIMZ,this%cNDIMT
  write(*,'("config:        NODEID:",I4)')  this%cnodeid
  write(*,'("config:     NODE XYZT:",4I3)') this%cipsite(1),this%cipsite(2),this%cipsite(3),this%cipsite(4)
  write(*,'("config:           PLQ:",E24.15)')  this%plq
  write(*,'("config:  Beta,cRG0,u0:",3E24.15)') this%beta,this%cRG0,this%u0
  write(*,'("config:   #PS-fermion:",I3)')      this%num_quark
  do iqk=1,this%num_quark
    write(*,'("config:quark(",I3,")%nf :",I3)')     iqk,this%quark(iqk)%nf
    write(*,'("config:quark(",I3,")%K  :",E24.15)') iqk,this%quark(iqk)%K
    write(*,'("config:quark(",I3,")%Csw:",E24.15)') iqk,this%quark(iqk)%Csw
  enddo
  write(*,'(80("="))')

  return
end subroutine

subroutine read_ddhmc_config(fhead,cprof,uu)
!
! read ddhmc config and convert normal-ordering config
!
  use file_tools_class
  character(len=*),        intent(in)    :: fhead
  type(config_prof_ddhmc), intent(inout) :: cprof
  complex(DP), intent(inout) :: uu(:,:,:,:,:,:,:)
  type(timer) :: etime
  type(gvf) :: u
  character(256) :: fname
  integer :: ipx,ipy,ipz,ipt
  integer :: ideo,ieo,mu,id,ic,jc
  integer :: idd,idt,idz,idy,idx
  integer :: ibx,iby,ibz,ibt,ieoxyz,ibth
  integer :: it,iz,iy,ix
  integer :: itt,itz,ity,itx
  integer :: iout

  call new(etime,master_only=.true.)
  call tic(etime)
  write(*,'("Read and Convert DDHMC measurement config to normal ordering.")')

  iout = search_free_file_unit()
  call new(u)

  do ipx=0,ddhmc_node%NDIMX-1
  do ipy=0,ddhmc_node%NDIMY-1
  do ipz=0,ddhmc_node%NDIMZ-1
  do ipt=0,ddhmc_node%NDIMT-1

    write(fname,'(A,".x",z1,"y",z1,"z",z1,"t",z1)')TRIM(ADJUSTL(fhead)),ipx,ipy,ipz,ipt
    write(*,'("Read: ",A)')TRIM(ADJUSTL(fname))
    open(iout,file=TRIM(fname),status='old',form='unformatted')
    do ideo=0,1
    do ieo=0,1
    do mu=1,NDIM
    do id=0,ddhmc_node%NDM-1
      read(iout) &
 &     u%deo(ideo)%eo(ieo)%mu(mu)%b(id)%s(0:ddhmc_node%NBTH,1:ddhmc_node%NBZ,1:ddhmc_node%NBY,1:ddhmc_node%NBX)
    enddo
    enddo
    enddo
    enddo
    call read_config_prof(cprof,iout)
    close(iout)
    call get_node_params(ddhmc_node,cprof%cipsite)

    do ieo =0,1
    do ideo=0,1
    do mu=1,NDIM
    do id = 0,ddhmc_node%NDM-1
      idd=id*2+ddhmc_node%idmod(id,ideo) ! calc. block index in 1-pu
      idt=mod(idd,ddhmc_node%NDT)
      idz=mod(idd/ddhmc_node%NDT,ddhmc_node%NDZ)
      idy=mod(idd/ddhmc_node%NDT/ddhmc_node%NDZ,ddhmc_node%NDY)
      idx=mod(idd/ddhmc_node%NDT/ddhmc_node%NDZ/ddhmc_node%NDY,ddhmc_node%NDX)

      do ibx=1,ddhmc_node%NBX  ! site loop in 1 block.
      do iby=1,ddhmc_node%NBY
      do ibz=1,ddhmc_node%NBZ
      ieoxyz=mod(ddhmc_node%iddeo(id,ideo)+ibz+iby+ibx+ieo,2)
      do ibth=1-ieoxyz,ddhmc_node%NBTH-ieoxyz
        ibt=ibth*2+ieoxyz
        ix=ibx+idx*ddhmc_node%NBX ! calc. global site index
        iy=iby+idy*ddhmc_node%NBY
        iz=ibz+idz*ddhmc_node%NBZ
        it=ibt+idt*ddhmc_node%NBT

        itx = ix + ipx*ddhmc_node%NX
        ity = iy + ipy*ddhmc_node%NY
        itz = iz + ipz*ddhmc_node%NZ
        itt = it + ipt*ddhmc_node%NT

        do jc=1,COL
        do ic=1,COL
          uu(ic,jc,itt,itz,ity,itx,mu) = u%deo(ideo)%eo(ieo)%mu(mu)%b(id)%s(ibth,ibz,iby,ibx)%u(ic,jc)
        enddo
        enddo

      enddo
      enddo
      enddo
      enddo
    enddo
    enddo
    enddo
    enddo

  enddo
  enddo
  enddo
  enddo
  call delete(u)

  call toc(etime)
  write(*,'(" done. ETIME:",F10.4)')get_elapse(etime)
  call delete(etime)

  return
end subroutine

subroutine plaquette(u,plq)
  complex(DP), intent(inout) :: u(:,:,:,:,:,:,:)
  real(DP),    intent(out)   :: plq
  complex(DP) :: uw0(COL,COL),uw1(COL,COL)
  real(DP) :: rtmp
  integer :: mu,nu,ic,jc
  integer :: itx,ity,itz,itt
  integer :: itx1,ity1,itz1,itt1
  integer :: itx2,ity2,itz2,itt2

  plq = 0.0_DP
  do mu=1,NDIM
  do nu=1,NDIM
    if (mu == nu) cycle
    rtmp = 0.0_DP
!$OMP PARALLEL DO PRIVATE(itx,ity,itz,itt, &
!$OMP&                    itx1,ity1,itz1,itt1, &
!$OMP&                    itx2,ity2,itz2,itt2, &
!$OMP&                    ic,jc,uw0,uw1) REDUCTION(+:rtmp)
    do itx=1,ddhmc_node%NTX
    do ity=1,ddhmc_node%NTY
    do itz=1,ddhmc_node%NTZ
    do itt=1,ddhmc_node%NTT
      itt1 = mod(itt + ist(mu)-1,ddhmc_node%NTT)+1
      itz1 = mod(itz + isz(mu)-1,ddhmc_node%NTZ)+1
      ity1 = mod(ity + isy(mu)-1,ddhmc_node%NTY)+1
      itx1 = mod(itx + isx(mu)-1,ddhmc_node%NTX)+1
      do jc=1,COL
      do ic=1,COL
        uw0(ic,jc) = u(ic,1,itt,itz,ity,itx,mu)*u(1,jc,itt1,itz1,ity1,itx1,nu) &
 &                 + u(ic,2,itt,itz,ity,itx,mu)*u(2,jc,itt1,itz1,ity1,itx1,nu) &
 &                 + u(ic,3,itt,itz,ity,itx,mu)*u(3,jc,itt1,itz1,ity1,itx1,nu)
      enddo
      enddo
      itt2 = mod(itt + ist(nu)-1,ddhmc_node%NTT)+1
      itz2 = mod(itz + isz(nu)-1,ddhmc_node%NTZ)+1
      ity2 = mod(ity + isy(nu)-1,ddhmc_node%NTY)+1
      itx2 = mod(itx + isx(nu)-1,ddhmc_node%NTX)+1
      do jc=1,COL
      do ic=1,COL
        uw1(ic,jc) = u(ic,1,itt,itz,ity,itx,nu)*u(1,jc,itt2,itz2,ity2,itx2,mu) &
 &                 + u(ic,2,itt,itz,ity,itx,nu)*u(2,jc,itt2,itz2,ity2,itx2,mu) &
 &                 + u(ic,3,itt,itz,ity,itx,nu)*u(3,jc,itt2,itz2,ity2,itx2,mu)
      enddo
      enddo
  
      rtmp = rtmp + REAL(uw0(1,1)*conjg(uw1(1,1))+uw0(1,2)*conjg(uw1(1,2))+uw0(1,3)*conjg(uw1(1,3)) &
 &                      +uw0(2,1)*conjg(uw1(2,1))+uw0(2,2)*conjg(uw1(2,2))+uw0(2,3)*conjg(uw1(2,3)) &
 &                      +uw0(3,1)*conjg(uw1(3,1))+uw0(3,2)*conjg(uw1(3,2))+uw0(3,3)*conjg(uw1(3,3)), kind=DP )
    enddo
    enddo
    enddo
    enddo
    plq = plq + rtmp
  enddo
  enddo

  plq = plq/COL/REAL(ddhmc_node%NTX*ddhmc_node%NTY*ddhmc_node%NTZ*ddhmc_node%NTT,kind=DP)/(NDIM*(NDIM-1))
  return
end subroutine

subroutine convert_and_save_config(fhead,u,plq)
!
! Convert normal config to e/o-config format
!
!  Splitting for parallel
! Reordering for even/odd
!
!  fhead : config file name
!      u : normal ordering config
!    plq : plaquette value
!
! When        fhead = "myconfig"
!   is given
!             myconfig.x?y?z?
!   is saved.
!
  use file_tools_class
  use field_gauge_class
  implicit none
  character(len=*), intent(in) :: fhead
  complex(DP),      intent(in) :: u(:,:,:,:,:,:,:)
  real(DP),         intent(in) :: plq

  character(len=CHARLEN) :: fnode,fname
#ifdef _ASYNC_IO_
  type(vfield_gluon_wg), asynchronous, allocatable :: uu
#else
  type(vfield_gluon_wg), allocatable :: uu
#endif

  integer :: io_list(0:NDIMX-1,0:NDIMY-1,0:NDIMZ-1)
  integer :: iout
  integer :: ipx,ipy,ipz
  integer :: mu,ix,iy,iz,it,ith,ic,jc
  integer :: itx,ity,itz,itt,ieo
  type(timer) :: etime


  write(*,'("outconfig: NODEDIM XYZ:",3I3)')NDIMX,NDIMY,NDIMZ
  !-----------------------------
  ! open parallel config files
  !-----------------------------
  do ipz=0,NDIMZ-1
  do ipy=0,NDIMY-1
  do ipx=0,NDIMX-1

    io_list(ipx,ipy,ipz) = search_free_file_unit()
    iout = io_list(ipx,ipy,ipz)

    write(fnode,'(".x",z1,"y",z1,"z",z1)')ipx,ipy,ipz
    fname=TRIM(fhead)//TRIM(fnode)
    write(*,'("Open:",A)')TRIM(fname)
#ifdef _ASYNC_IO_
    open(unit=iout,file=fname,form='unformatted',status='unknown',asynchronous='yes')
#else
    open(unit=iout,file=fname,form='unformatted',status='unknown')
#endif

  enddo
  enddo
  enddo

  write(*,'("Converting normal ordering to our ordering....")',advance='no')

  call new(etime,master_only=.true.)
  call tic(etime)

  allocate(uu)
  call new(uu)
  call clear(uu)

  !-----------------------------
  ! splitting, reordering
  !              and storing
  !-----------------------------
  do ipz=0,NDIMZ-1
  do ipy=0,NDIMY-1
  do ipx=0,NDIMX-1

    do mu=1,NDIM
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,itx,ity,itz,itt,ith,ieo,jc,ic)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
    do it=1,NT
      itx = ix + ipx*NX
      ity = iy + ipy*NY
      itz = iz + ipz*NZ
      itt = it
      ith = it/2
      ieo = mod(itt+itz+ity+itx,2)
      do jc=1,COL
      do ic=1,COL
        uu%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,jc)  &
 &                         = u(ic,jc,itt,itz,ity,itx,mu)
      enddo
      enddo
    enddo
    enddo
    enddo
    enddo
    enddo

    iout = io_list(ipx,ipy,ipz)
#ifdef _ASYNC_IO_
    write(unit=iout,asynchronous='yes') uu
    write(unit=iout,asynchronous='yes') NTX,NTY,NTZ,NTT
    write(unit=iout,asynchronous='yes') NDIMX,NDIMY,NDIMZ
    write(unit=iout,asynchronous='yes') plq
#else
    write(unit=iout) uu
    write(unit=iout) NTX,NTY,NTZ,NTT
    write(unit=iout) NDIMX,NDIMY,NDIMZ
    write(unit=iout) plq
#endif

  enddo
  enddo
  enddo

  deallocate(uu)

  call toc(etime)
  write(*,'(" done. ETIME:",F10.4)')get_elapse(etime)
  call delete(etime)

  write(*,'("Waiting for Storing data....")',advance='no')

  call new(etime,master_only=.true.)
  call tic(etime)

  do ipz=0,NDIMZ-1
  do ipy=0,NDIMY-1
  do ipx=0,NDIMX-1
    iout = io_list(ipx,ipy,ipz)
    close(unit=iout)
  enddo
  enddo
  enddo

  call toc(etime)
  write(*,'(" done. ETIME:",F10.4)')get_elapse(etime)
  call delete(etime)

  return
end subroutine

end module
