#include <config.h>

!
!= Definitions of gauge fields
!
!== Version
!
! $Id: field_gauge_class.F90,v 1.33 2011/10/30 02:41:31 ishikawa Exp $
!
module field_gauge_class
!
!= This defines even/odd site orderd gauge fields and related routines.
!
! Operations on fields are in linear algebra and are expressed with subroutine form.
! The subroutine name contains the fllowing strings expressing operations for the input arguments.
!
![<tt>\assign</tt>] substitution or data copy
![<tt>\accum</tt>]  accumulation to first argument
![<tt>\conjg</tt>]  take Hermitian conjugate
![<tt>\mult</tt>]   multiplication of scalar on field or multiplication of two gauge fields
![<tt>\add</tt>]    addition of two fields
![<tt>\sub</tt>]    subtractoin of two fields
![<tt>\shift</tt>]  shift cordinate of the field
![<tt>\abs2</tt>]   squared norm of a field.
! etc...
!
! ---------
!
! The subroutine name can contains several operations to do composite opertions.
! The order of subrutine arguments and operations in subroutine name is matched as fllows.
!
![<tt>\assign_add(u,v,w)</tt>] 
!                                          u \assign v \add w           : u <= v + w 
![<tt>\assign_sub(u,v,w)</tt>] 
!                                          u \assign v \sub w           : u <= v - w 
![<tt>\accum_sub(u,v)</tt>] 
!                                          u \accum u \sub v            : u <= u - v
![<tt>\assign_mult(u,v,w)</tt>] 
!                                          u \assign v \mult w          : u <= v * w
![<tt>\assign_shift(u,v,mu)</tt>] 
!                                          u(n) \assign shift(v,mu)     : u(n) <= v(n+mu)
! etc...
!
  use lattice_class
  use timer_class
  use error_class
  use comlib
  implicit none
  private

  public :: su3fm

  public :: sfield_gluon_eo_wg
  public :: sfield_gluon_eo_wog
  public :: sfield_gluon_wg
  public :: sfield_gluon_wog

  public :: vfield_gluon_eo_wg
  public :: vfield_gluon_eo_wog
  public :: vfield_gluon_wg
  public :: vfield_gluon_wog

  public :: tfield_gluon_eo_wg
  public :: tfield_gluon_wg
  
  public :: copy_sfg_time
  public :: print_copy_sfg_statistics

  type(timer), save :: copy_sfg_time               ! holds gauge field boundary copy timing
  logical,     save :: m_is_initialized = .FALSE.  ! holds gauge field boundary copy buffer initialization status

  !
  ! Order of Taylor-expansion approximation for su(3) matrix exponential 
  !
  integer, parameter :: EXP_TAYLOR_ORDER = 14

  !
  !  boundary copy related variables
  !
  integer, parameter :: GBUFF_SIZE(2*(NDIM-1))  &
 &                                =(/COL**2*(NTH+1)*NZ*NY,    &! for x-direction w/o cornar
 &                                   COL**2*(NTH+1)*NZ*NX,    &! for y-direction w/o cornar
 &                                   COL**2*(NTH+1)*NY*NX,    &! for z-direction w/o cornar
 &                                   COL**2*(NTH+1)*NZ*2,     &! for x-direction at cornar (x-y plane)
 &                                   COL**2*(NTH+1)*NY*2,     &! for x-direction at cornar (x-z plane)
 &                                   COL**2*(NTH+1)*NX*2  /)   ! for y-direction at cornar (y-z plane)
  integer, parameter :: GBUFF_MAX_SIZE=MAX(GBUFF_SIZE(1), &
 &                                         GBUFF_SIZE(2), &
 &                                         GBUFF_SIZE(3), &
 &                                         GBUFF_SIZE(4), &
 &                                         GBUFF_SIZE(5), &
 &                                         GBUFF_SIZE(6))  
  complex(DP), save, target :: gbuffup(GBUFF_MAX_SIZE,2,2*(NDIM-1)) ! gauge field ghost sites buffer
  complex(DP), save, target :: gbuffdn(GBUFF_MAX_SIZE,2,2*(NDIM-1)) ! gauge field ghost sites buffer
  type(comlib_data), save   :: idgsendup(2*(NDIM-1))
  type(comlib_data), save   :: idgsenddn(2*(NDIM-1))  ! comlib data for ghost site exchange.

  type :: fields_gauge
  !
  ! gauge field initialization status
  !
    integer, private :: idummy(2) = 0
  end type

  type su3fm
  !
  ! su(3) matrix
  !
    complex(DP) :: u(COL,COL)
  end type

  type sfield_gluon_eo_wg
  !
  ! su(3) matrix scalar field on even/odd sites with ghost sits
  ! used for gauge vector/tensor fields.
  !
    type(su3fm) :: s(0:NTH,0:NZ1,0:NY1,0:NX1)
    integer :: ieo,idummy(3)
  end type

  type sfield_gluon_eo_wog
  !
  ! su(3) matrix scalar field on even/odd sites without ghost sits.
  ! used for gauge mementum vector field.
  !
    type(su3fm) :: s(NTH,NZ,NY,NX)
    integer :: ieo,idummy(3)
  end type

  type vfield_gluon_eo_wg
  !
  ! su(3) matrix vector field on even/odd sites with ghost sits
  !
    type(sfield_gluon_eo_wg)  :: mu(NDIM)
    integer :: ieo,idummy(3)
  end type

  type vfield_gluon_eo_wog
  !
  ! su(3) matrix vector field on even/odd sites without ghost sits
  !
    type(sfield_gluon_eo_wog)  :: mu(NDIM)
    integer :: ieo,idummy(3)
  end type

  type tfield_gluon_eo_wg
  !
  ! su(3) matrix symmetric tensor field on even/odd sites with ghost sits
  !
    type(sfield_gluon_eo_wg)  :: munu(NDIM*(NDIM-1)/2)
    integer :: ieo,idummy(3)
  end type

  type sfield_gluon_wg
  !
  ! su(3) matrix scalar field with ghost sits
  !
    type(sfield_gluon_eo_wg)  :: eo(0:1)
  end type

  type sfield_gluon_wog
  !
  ! su(3) matrix scalar field without ghost sits
  !
    type(sfield_gluon_eo_wog)  :: eo(0:1)
  end type

  type vfield_gluon_wg
  !
  ! su(3) matrix vector field with ghost sits
  !
    type(vfield_gluon_eo_wg)  :: eo(0:1)
  end type

  type vfield_gluon_wog
  !
  ! su(3) matrix vector field without ghost sits
  !
    type(vfield_gluon_eo_wog)  :: eo(0:1)
  end type

  type tfield_gluon_wg
  !
  ! su(3) matrix symmetric tensor field with ghost sits
  !
    type(tfield_gluon_eo_wg)  :: eo(0:1)
  end type

  public :: copy_boundary

  interface copy_boundary
  !
  ! gauge scalr field e/o ghost site copy
  !
    module procedure  copy_sfg_eo
    module procedure  copy_sfg
    module procedure  copy_vfg_eo
    module procedure  copy_vfg
    module procedure  copy_tfg_eo
    module procedure  copy_tfg
  end interface

  public :: new
  public :: clear
  public :: assign
  public :: assign_shift
  public :: assign_conjg
  public :: assign_conjg_shift
  public :: assign_mult
  public :: assign_conjg1_mult
  public :: assign_conjg2_mult
  public :: accum_add
  public :: accum_sub
  public :: assign_add
  public :: assign_sub
  public :: trace
  public :: abs2
  public :: traceless_antihermite
  public :: accum_add_mult

  public :: set_identity
  public :: set_gaussian_noise
  public :: plaquette
  public :: update_u
  public :: update_p
  public :: get_force_norm
  public :: make_two_links

  interface new
  !
  ! field initialization (assin even/odd index)
  !
    module procedure new_sfg_eo_wg
    module procedure new_sfg_eo_wog
    module procedure new_sfg_wg
    module procedure new_sfg_wog

    module procedure new_vfg_eo_wg
    module procedure new_vfg_eo_wog
    module procedure new_vfg_wg
    module procedure new_vfg_wog

    module procedure new_tfg_eo_wg
    module procedure new_tfg_wg
  end interface


  interface clear
  !
  ! set zeros 
  !
    module procedure clear_sfg_eo_wg
    module procedure clear_sfg_eo_wog

    module procedure clear_vfg_eo_wg
    module procedure clear_vfg_wg

    module procedure clear_vfg_eo_wog
    module procedure clear_vfg_wog

    module procedure clear_tfg_eo_wg
    module procedure clear_tfg_wg
  end interface

  interface assign
  !
  ! assign
  !
    module procedure assign_sfg_eo_wg
  end interface

  interface assign_shift
  !
  ! shift field site index by 1 site in 1dim or 2dim directions.
  !
    module procedure assign_shift1_sfg_eo_wg
    module procedure assign_shift2_sfg_eo_wg
  end interface

  interface assign_conjg
  !
  ! take Hermitian conjugate 
  !
    module procedure assign_conjg_sfg_eo_wg
  end interface

  interface assign_conjg_shift
  !
  ! take dagger conjugate and 
  ! shift field site index by 1 site in 1dim or 2dim directions.
  !
    module procedure assign_conjg_shift1_sfg_eo_wg
    module procedure assign_conjg_shift2_sfg_eo_wg
  end interface

  interface assign_mult
  !
  ! multiply two fields or multiply scalar
  !
    module procedure  assign_mult_vfg_wg_wog
    module procedure  assign_mult_sfg_eo_wg_wog
    module procedure  assign_mult_sfg_eo_wg
    module procedure  assign_rmult_sfg_eo_wg
  end interface

  interface accum_add_mult
  !
  ! multiply scaler on field and accumulate
  !
  ! u <= u + v * coef
  !
    module procedure accum_add_cmult_sfg_eo_wog
  end interface

  interface assign_conjg1_mult
  !
  ! multiply Hermitian conjugated field and normal field
  !
  ! u <= v^+ * w
  !
    module procedure assign_conjg1_mult_sfg_eo_wg
  end interface

  interface assign_conjg2_mult
  !
  ! multiply normal field and Hermitian conjugated field
  !
  ! u <= w * v^+
  !
    module procedure assign_conjg2_mult_sfg_eo_wg
  end interface

  interface accum_add
  !
  ! u <= u + v 
  !
    module procedure accum_add_sfg_eo_wg
  end interface

  interface accum_sub
  !
  ! u <= u - v 
  !
    module procedure accum_sub_sfg_eo_wg
  end interface

  interface assign_add
  !
  ! u <= v + w 
  !
    module procedure assign_add_sfg_eo_wg
  end interface

  interface assign_sub
  !
  ! u <= v - w 
  !
    module procedure assign_sub_sfg_eo_wg
  end interface

  interface trace
  !
  ! take trace
  !
    module procedure trace_sfg_eo_wg
  end interface

  interface traceless_antihermite
    !
    ! extract traceless and anti Hermitian part
    !
    module procedure traceless_ahermite_sfg_eo_wog
    module procedure traceless_ahermite_vfg_wog
  end interface

  interface abs2
    !
    ! squared absolute value
    !
    module procedure abs2_sfg_eo_wog
  end interface

  interface set_identity
    module procedure set_identity_u
  end interface

  interface set_gaussian_noise
    module procedure set_gaussian_noise_p
  end interface

  interface plaquette
    module procedure plaquette_u
  end interface

  interface update_p
    module procedure update_p_wog
  end interface

  interface get_force_norm
    module procedure get_force_norm_wog
  end interface

!=================================
! SF boundary condition related
!=================================
  public :: set_sf_boundary
  public :: accum_mult_at

  interface set_sf_boundary
  !
  ! set SF boundary condition
  !
    module procedure set_sf_boundary_u
    module procedure set_sf_boundary_u_eo
    module procedure set_sf_boundary_p
    module procedure set_sf_boundary_p_eo
  end interface

  interface accum_mult_at
  !
  ! multiply a real constant on a SU(3) scalar field 
  ! at time slice t0, (or t0 and t1)
  !
  !  U(t0,z,y,x) = U(t0,z,y,x) * coef
  !  U(t1,z,y,x) = U(t1,z,y,x) * coef   (optional)
  !
    module procedure accum_mult_at_sf_boundary_u
    module procedure accum_mult_at_sf_boundary_u_eo
  end interface


contains

subroutine print_copy_sfg_statistics()
  implicit none
  integer :: mu
  if (.not.m_is_initialized) return
  if (0 == nodeid) then
    do mu=1,NDIM*(NDIM-1)/2
      call comlib_print_statistics(idgsendup(mu))
      call comlib_print_statistics(idgsenddn(mu))
    enddo
  endif
  return
end subroutine

subroutine copy_sfg_eo(ue)
!
! Boundary copy for su(3) scalar field with periodic boundary condition.
!
!  ue : even/odd site su3 scalar field (ieo=0/1)
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: ue

  type(fields_gauge) :: fields
  integer :: ieoxyz,itb0,itb1
  integer :: ix,iy,iz,itb
#ifndef _singlePU
  integer :: ibuff,ic,jc
#endif

  if (.not.m_is_initialized) call new_fields_gauge(fields)

  call tic(copy_sfg_time)
#ifndef _singlePU
  call comlib_barrier
#endif

!****** T - boundaray
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb0,itb1)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ue%ieo,2)
    itb0=   ieoxyz *NTH
    itb1=(1-ieoxyz)*NTH
    ue%s(itb0,iz,iy,ix)=ue%s(itb1,iz,iy,ix)
  enddo
  enddo
  enddo

#ifndef _singlePU
!----------------------------------------------------------
! copy without cornar
! set send buffer
!----------------------------------------------------------
!****** Z - boundaray
#if _NDIMZ == 1
!$OMP PARALLEL DO PRIVATE(ix,iy,itb)
  do ix=1,NX
  do iy=1,NY
    do itb=0,NTH
      ue%s(itb,  0,iy,ix)=ue%s(itb,NZ,iy,ix)
      ue%s(itb,NZ1,iy,ix)=ue%s(itb, 1,iy,ix)
    enddo
  enddo
  enddo
#else
  ibuff=0
  do ix=1,NX
  do iy=1,NY
    do itb=0,NTH
      do jc=1,COL
      do ic=1,COL
        ibuff=ibuff+1
        gbuffup(ibuff,1,3)=ue%s(itb,NZ,iy,ix)%u(ic,jc)
        gbuffdn(ibuff,1,3)=ue%s(itb, 1,iy,ix)%u(ic,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  if (ibuff.NE.GBUFF_SIZE(3)) then
    call error_stop(" Error : com-buff size in Z")
  endif
#endif

!****** Y - boundaray
#if _NDIMY == 1
!$OMP PARALLEL DO PRIVATE(ix,iz,itb)
  do ix=1,NX
  do iz=1,NZ
    do itb=0,NTH
      ue%s(itb,iz,  0,ix)=ue%s(itb,iz,NY,ix)
      ue%s(itb,iz,NY1,ix)=ue%s(itb,iz, 1,ix)
    enddo
  enddo
  enddo
#else
  ibuff=0
  do ix=1,NX
  do iz=1,NZ
    do itb=0,NTH
      do jc=1,COL
      do ic=1,COL
        ibuff=ibuff+1
        gbuffup(ibuff,1,2)=ue%s(itb,iz,NY,ix)%u(ic,jc)
        gbuffdn(ibuff,1,2)=ue%s(itb,iz, 1,ix)%u(ic,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  if (ibuff.NE.GBUFF_SIZE(2)) then
    call error_stop(" Error : com-buff size in Y")
  endif
#endif

!****** X - boundaray
#if _NDIMX == 1
!$OMP PARALLEL DO PRIVATE(iy,iz,itb)
  do iy=1,NY
  do iz=1,NZ
    do itb=0,NTH
      ue%s(itb,iz,iy,  0)=ue%s(itb,iz,iy,NX)
      ue%s(itb,iz,iy,NX1)=ue%s(itb,iz,iy, 1)
    enddo
  enddo
  enddo
#else
  ibuff=0
  do iy=1,NY
  do iz=1,NZ
    do itb=0,NTH
      do jc=1,COL
      do ic=1,COL
        ibuff=ibuff+1
        gbuffup(ibuff,1,1)=ue%s(itb,iz,iy,NX)%u(ic,jc)
        gbuffdn(ibuff,1,1)=ue%s(itb,iz,iy, 1)%u(ic,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  if (ibuff.NE.GBUFF_SIZE(1)) then
    call error_stop(" Error : com-buff size in X")
  endif
#endif

#if _NDIMZ != 1
  call comlib_sendrecv(idgsendup(3))
  call comlib_sendrecv(idgsenddn(3))
#endif
#if _NDIMY != 1
  call comlib_sendrecv(idgsendup(2))
  call comlib_sendrecv(idgsenddn(2))
#endif
#if _NDIMX != 1
  call comlib_sendrecv(idgsendup(1))
  call comlib_sendrecv(idgsenddn(1))
#endif

!****** Z - boundaray
#if _NDIMZ != 1
  ibuff=0
  do ix=1,NX
  do iy=1,NY
    do itb=0,NTH
      do jc=1,COL
      do ic=1,COL
        ibuff=ibuff+1
        ue%s(itb,  0,iy,ix)%u(ic,jc)=gbuffup(ibuff,2,3)
        ue%s(itb,NZ1,iy,ix)%u(ic,jc)=gbuffdn(ibuff,2,3)
      enddo
      enddo
    enddo
  enddo
  enddo
#endif

!****** Y - boundaray
#if _NDIMY != 1
  ibuff=0
  do ix=1,NX
  do iz=1,NZ
    do itb=0,NTH
      do jc=1,COL
      do ic=1,COL
        ibuff=ibuff+1
        ue%s(itb,iz,  0,ix)%u(ic,jc)=gbuffup(ibuff,2,2)
        ue%s(itb,iz,NY1,ix)%u(ic,jc)=gbuffdn(ibuff,2,2)
      enddo
      enddo
    enddo
  enddo
  enddo
#endif

!****** X - boundaray
#if _NDIMX != 1
  ibuff=0
  do iy=1,NY
  do iz=1,NZ
    do itb=0,NTH
      do jc=1,COL
      do ic=1,COL
        ibuff=ibuff+1
        ue%s(itb,iz,iy,  0)%u(ic,jc)=gbuffup(ibuff,2,1)
        ue%s(itb,iz,iy,NX1)%u(ic,jc)=gbuffdn(ibuff,2,1)
      enddo
      enddo
    enddo
  enddo
  enddo
#endif

!--------------------------------------------------------
! copy cornar
!--------------------------------------------------------
  call comlib_barrier

!****** T - boundaray
!****** t-x
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb0,itb1)
  do ix=0,NX1,NX1
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ue%ieo,2)
    itb0=   ieoxyz *NTH
    itb1=(1-ieoxyz)*NTH
    ue%s(itb0,iz,iy,ix)=ue%s(itb1,iz,iy,ix)
  enddo
  enddo
  enddo
!****** t-y
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb0,itb1)
  do ix=1,NX
  do iy=0,NY1,NY1
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ue%ieo,2)
    itb0=   ieoxyz *NTH
    itb1=(1-ieoxyz)*NTH
    ue%s(itb0,iz,iy,ix)=ue%s(itb1,iz,iy,ix)
  enddo
  enddo
  enddo
!****** t-z
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb0,itb1)
  do ix=1,NX
  do iy=1,NY
  do iz=0,NZ1,NZ1
    ieoxyz=mod(ipeo+ix+iy+iz+ue%ieo,2)
    itb0=   ieoxyz *NTH
    itb1=(1-ieoxyz)*NTH
    ue%s(itb0,iz,iy,ix)=ue%s(itb1,iz,iy,ix)
  enddo
  enddo
  enddo

!*********************** copy cornar
!*********************** send
!****** x-y
#if _NDIMX == 1
!$OMP PARALLEL DO PRIVATE(iz,itb)
    do iz=1,NZ
      do itb=0,NTH
        ue%s(itb,iz,  0,  0)=ue%s(itb,iz,  0,NX)
        ue%s(itb,iz,NY1,  0)=ue%s(itb,iz,NY1,NX)
        ue%s(itb,iz,  0,NX1)=ue%s(itb,iz,  0, 1)
        ue%s(itb,iz,NY1,NX1)=ue%s(itb,iz,NY1, 1)
      enddo
    enddo
#else
    ibuff=0
    do iz=1,NZ
      do itb=0,NTH
        do jc=1,COL
        do ic=1,COL
          ibuff=ibuff+2
          gbuffup(ibuff-1,1,4)=ue%s(itb,iz,  0,NX)%u(ic,jc)
          gbuffup(ibuff  ,1,4)=ue%s(itb,iz,NY1,NX)%u(ic,jc)
          gbuffdn(ibuff-1,1,4)=ue%s(itb,iz,  0, 1)%u(ic,jc)
          gbuffdn(ibuff  ,1,4)=ue%s(itb,iz,NY1, 1)%u(ic,jc)
        enddo
        enddo
      enddo
    enddo
    if (ibuff.NE.GBUFF_SIZE(4)) then
      call error_stop(" Error : com-buff size in (x-y)")
    endif
#endif
!****** x-z
#if _NDIMX == 1
!$OMP PARALLEL DO PRIVATE(iy,itb)
    do iy=1,NY
      do itb=0,NTH
        ue%s(itb,  0,iy,  0)=ue%s(itb,  0,iy,NX)
        ue%s(itb,NZ1,iy,  0)=ue%s(itb,NZ1,iy,NX)
        ue%s(itb,  0,iy,NX1)=ue%s(itb,  0,iy, 1)
        ue%s(itb,NZ1,iy,NX1)=ue%s(itb,NZ1,iy, 1)
      enddo
    enddo
#else
    ibuff=0
    do iy=1,NY
      do itb=0,NTH
        do jc=1,COL
        do ic=1,COL
          ibuff=ibuff+2
          gbuffup(ibuff-1,1,5)=ue%s(itb,  0,iy,NX)%u(ic,jc)
          gbuffup(ibuff  ,1,5)=ue%s(itb,NZ1,iy,NX)%u(ic,jc)
          gbuffdn(ibuff-1,1,5)=ue%s(itb,  0,iy, 1)%u(ic,jc)
          gbuffdn(ibuff  ,1,5)=ue%s(itb,NZ1,iy, 1)%u(ic,jc)
        enddo
        enddo
      enddo
    enddo
    if (ibuff.NE.GBUFF_SIZE(5)) then
      call error_stop(" Error : com-buff size in (x-z)")
    endif
#endif
!****** y-z
#if _NDIMY == 1
!$OMP PARALLEL DO PRIVATE(ix,itb)
    do ix=1,NX
      do itb=0,NTH
        ue%s(itb,  0,  0,ix)=ue%s(itb,  0,NY,ix)
        ue%s(itb,NZ1,  0,ix)=ue%s(itb,NZ1,NY,ix)
        ue%s(itb,  0,NY1,ix)=ue%s(itb,  0, 1,ix)
        ue%s(itb,NZ1,NY1,ix)=ue%s(itb,NZ1, 1,ix)
      enddo
    enddo
#else
    ibuff=0
    do ix=1,NX
      do itb=0,NTH
        do jc=1,COL
        do ic=1,COL
          ibuff=ibuff+2
          gbuffup(ibuff-1,1,6)=ue%s(itb,  0,NY,ix)%u(ic,jc)
          gbuffup(ibuff  ,1,6)=ue%s(itb,NZ1,NY,ix)%u(ic,jc)
          gbuffdn(ibuff-1,1,6)=ue%s(itb,  0, 1,ix)%u(ic,jc)
          gbuffdn(ibuff  ,1,6)=ue%s(itb,NZ1, 1,ix)%u(ic,jc)
        enddo
        enddo
      enddo
    enddo
    if (ibuff.NE.GBUFF_SIZE(6)) then
      call error_stop(" Error : com-buff size in (y-z)")
    endif
#endif

#if _NDIMX != 1
    call comlib_sendrecv(idgsendup(4))
    call comlib_sendrecv(idgsenddn(4))
    call comlib_sendrecv(idgsendup(5))
    call comlib_sendrecv(idgsenddn(5))
#endif
#if _NDIMY != 1
    call comlib_sendrecv(idgsendup(6))
    call comlib_sendrecv(idgsenddn(6))
#endif

!*********************** recv
!****** x-y
#if _NDIMX != 1
    ibuff=0
    do iz=1,NZ
      do itb=0,NTH
        do jc=1,COL
        do ic=1,COL
          ibuff=ibuff+2
          ue%s(itb,iz,  0,  0)%u(ic,jc)=gbuffup(ibuff-1,2,4)
          ue%s(itb,iz,NY1,  0)%u(ic,jc)=gbuffup(ibuff  ,2,4)
          ue%s(itb,iz,  0,NX1)%u(ic,jc)=gbuffdn(ibuff-1,2,4)
          ue%s(itb,iz,NY1,NX1)%u(ic,jc)=gbuffdn(ibuff  ,2,4)
        enddo
        enddo
      enddo
    enddo
!****** x-z
    ibuff=0
    do iy=1,NY
      do itb=0,NTH
        do jc=1,COL
        do ic=1,COL
          ibuff=ibuff+2
          ue%s(itb,  0,iy,  0)%u(ic,jc)=gbuffup(ibuff-1,2,5)
          ue%s(itb,NZ1,iy,  0)%u(ic,jc)=gbuffup(ibuff  ,2,5)
          ue%s(itb,  0,iy,NX1)%u(ic,jc)=gbuffdn(ibuff-1,2,5)
          ue%s(itb,NZ1,iy,NX1)%u(ic,jc)=gbuffdn(ibuff  ,2,5)
        enddo
        enddo
      enddo
    enddo
#endif
!****** y-z
#if _NDIMY != 1
    ibuff=0
    do ix=1,NX
      do itb=0,NTH
        do jc=1,COL
        do ic=1,COL
          ibuff=ibuff+2
          ue%s(itb,  0,  0,ix)%u(ic,jc)=gbuffup(ibuff-1,2,6)
          ue%s(itb,NZ1,  0,ix)%u(ic,jc)=gbuffup(ibuff  ,2,6)
          ue%s(itb,  0,NY1,ix)%u(ic,jc)=gbuffdn(ibuff-1,2,6)
          ue%s(itb,NZ1,NY1,ix)%u(ic,jc)=gbuffdn(ibuff  ,2,6)
        enddo
        enddo
      enddo
    enddo
#endif

#else

!************************************************************************
!****** Z - boundaray
!$OMP PARALLEL DO  &
!$OMP PRIVATE(ix,iy,itb)
  do ix = 1,NX
  do iy = 1,NY
    do itb = 0,NTH
      ue%s(itb,  0,iy,ix)=ue%s(itb,NZ,iy,ix)
      ue%s(itb,NZ1,iy,ix)=ue%s(itb, 1,iy,ix)
    enddo
  enddo
  enddo

!****** Y - boundaray
!$OMP PARALLEL DO  &
!$OMP PRIVATE(ix,iz,itb)
  do ix = 1,NX
  do iz = 0,NZ1
    do itb = 0,NTH
      ue%s(itb,iz,  0,ix)=ue%s(itb,iz,NY,ix)
      ue%s(itb,iz,NY1,ix)=ue%s(itb,iz, 1,ix)
    enddo
  enddo
  enddo

!****** X - boundaray
!$OMP PARALLEL DO  &
!$OMP COLLAPSE(2)  &
!$OMP PRIVATE(iy,iz,itb)
  do iy = 0,NY1
  do iz = 0,NZ1
    do itb = 0,NTH
      ue%s(itb,iz,iy,  0)=ue%s(itb,iz,iy,NX)
      ue%s(itb,iz,iy,NX1)=ue%s(itb,iz,iy, 1)
    enddo
  enddo
  enddo

#endif

  call toc(copy_sfg_time)

  return
end subroutine

subroutine copy_sfg(u)
  implicit none
  type(sfield_gluon_wg), intent(inout) :: u
  call copy_boundary(u%eo(0))
  call copy_boundary(u%eo(1))
  return
end subroutine

subroutine copy_vfg_eo(ue)
  implicit none
  type(vfield_gluon_eo_wg), intent(inout) :: ue
  integer :: mu
  do mu=1,NDIM
    call copy_boundary(ue%mu(mu))
  enddo
#ifdef _SF
  call set_sf_boundary(ue)
#endif
  return
end subroutine

subroutine copy_vfg(u)
  implicit none
  type(vfield_gluon_wg), intent(inout) :: u
  call copy_boundary(u%eo(0))
  call copy_boundary(u%eo(1))
  return
end subroutine

subroutine copy_tfg_eo(ue)
  implicit none
  type(tfield_gluon_eo_wg), intent(inout) :: ue
  integer :: munu
  do munu=1,NDIM*(NDIM-1)/2
    call copy_boundary(ue%munu(munu))
  enddo
  return
end subroutine

subroutine copy_tfg(u)
  implicit none
  type(tfield_gluon_wg), intent(inout) :: u
  call copy_boundary(u%eo(0))
  call copy_boundary(u%eo(1))
  return
end subroutine


subroutine new_sfg_eo_wg(ue,ieo)
!
! set even/odd index on the field 
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: ue
  integer, intent(in) :: ieo
  ue%ieo=ieo
  return
end subroutine

subroutine new_sfg_eo_wog(ue,ieo)
!
! set even/odd index on the field 
!
  implicit none
  type(sfield_gluon_eo_wog), intent(inout) :: ue
  integer, intent(in) :: ieo
  ue%ieo=ieo
  return
end subroutine

subroutine new_vfg_eo_wg(ue,ieo)
  !
  ! set even/odd index on the field 
  !
  implicit none
  type(vfield_gluon_eo_wg), intent(inout) :: ue
  integer, intent(in) :: ieo
  integer :: mu
  ue%ieo=ieo
  do mu=1,NDIM
    call new(ue%mu(mu),ieo)
  enddo
  return
end subroutine

subroutine new_vfg_eo_wog(ue,ieo)
  !
  ! set even/odd index on the field 
  !
  implicit none
  type(vfield_gluon_eo_wog), intent(inout) :: ue
  integer, intent(in) :: ieo
  integer :: mu
  ue%ieo=ieo
  do mu=1,NDIM
    call new(ue%mu(mu),ieo)
  enddo
  return
end subroutine

subroutine new_tfg_eo_wg(ue,ieo)
  !
  ! set even/odd index on the field 
  !
  implicit none
  type(tfield_gluon_eo_wg), intent(inout) :: ue
  integer, intent(in) :: ieo
  integer :: mu
  ue%ieo=ieo
  do mu=1,NDIM*(NDIM-1)/2
    call new(ue%munu(mu),ieo)
  enddo
  return
end subroutine

subroutine new_sfg_wg(u)
  !
  ! set even/odd index on the field 
  !
  implicit none
  type(sfield_gluon_wg), intent(inout) :: u
  call new(u%eo(0),0)
  call new(u%eo(1),1)
  return
end subroutine

subroutine new_sfg_wog(u)
  !
  ! set even/odd index on the field 
  !
  implicit none
  type(sfield_gluon_wog), intent(inout) :: u
  call new(u%eo(0),0)
  call new(u%eo(1),1)
  return
end subroutine

subroutine new_vfg_wg(u)
  !
  ! set even/odd index on the field 
  !
  implicit none
  type(vfield_gluon_wg), intent(inout) :: u
  call new(u%eo(0),0)
  call new(u%eo(1),1)
  return
end subroutine

subroutine new_vfg_wog(u)
  !
  ! set even/odd index on the field 
  !
  implicit none
  type(vfield_gluon_wog), intent(inout) :: u
  call new(u%eo(0),0)
  call new(u%eo(1),1)
  return
end subroutine

subroutine new_tfg_wg(u)
  !
  ! set even/odd index on the field 
  !
  implicit none
  type(tfield_gluon_wg), intent(inout) :: u
  call new(u%eo(0),0)
  call new(u%eo(1),1)
  return
end subroutine

subroutine assign_sfg_eo_wg(u2e,u1e)
  !
  !  u2e <= u1e
  !
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u2e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e
  integer :: ix,iy,iz,ieoxyz,itb
  u2e%ieo=u1e%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+u2e%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      u2e%s(itb,iz,iy,ix)=u1e%s(itb,iz,iy,ix)
    enddo
  enddo
  enddo
  enddo
  return
end subroutine

subroutine assign_shift1_sfg_eo_wg(u1o,u1e,mu)
!
!  u1o(n) <= u1e(n+mu)
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u1o
  type(sfield_gluon_eo_wg), intent(in)    :: u1e
  integer, intent(in) :: mu
  integer :: ix,iy,iz,ieoxyz,itb
  integer :: ix2,iy2,iz2,itb2,ic,jc

  u1o%ieo=1-u1e%ieo
  select case (mu)
  case (1:NDIM)
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u1o%ieo,2)
      ix2=ix+isx(mu)
      iy2=iy+isy(mu)
      iz2=iz+isz(mu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb+ieoxyz*ist(mu)
        u1o%s(itb,iz,iy,ix)=u1e%s(itb2,iz2,iy2,ix2)
      enddo
    enddo
    enddo
    enddo
  case (-NDIM:-1)
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u1o%ieo,2)
      ix2=ix-isx(-mu)
      iy2=iy-isy(-mu)
      iz2=iz-isz(-mu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb-(1-ieoxyz)*ist(-mu)
        u1o%s(itb,iz,iy,ix)=u1e%s(itb2,iz2,iy2,ix2)
      enddo
    enddo
    enddo
    enddo
  case default
    call error_stop("error with |mu|>NDIM in assign_shift1_sfg_eo_wg.")
  end select

  return
end subroutine

subroutine assign_conjg_shift1_sfg_eo_wg(u1o,u1e,mu)
!
!  u1o(n) <= (u1e(n+mu))^+
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u1o
  type(sfield_gluon_eo_wg), intent(in)    :: u1e
  integer, intent(in) :: mu
  integer :: ix,iy,iz,ieoxyz,itb
  integer :: ix2,iy2,iz2,itb2,ic,jc

  u1o%ieo=1-u1e%ieo
  select case (mu)
  case (1:NDIM)
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2,ic,jc)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u1o%ieo,2)
      ix2=ix+isx(mu)
      iy2=iy+isy(mu)
      iz2=iz+isz(mu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb+ieoxyz*ist(mu)
        do jc=1,COL
        do ic=1,COL
          u1o%s(itb,iz,iy,ix)%u(ic,jc)=conjg(u1e%s(itb2,iz2,iy2,ix2)%u(jc,ic))
        enddo
        enddo
      enddo
    enddo
    enddo
    enddo
  case (-NDIM:-1)
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2,ic,jc)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u1o%ieo,2)
      ix2=ix-isx(-mu)
      iy2=iy-isy(-mu)
      iz2=iz-isz(-mu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb-(1-ieoxyz)*ist(-mu)
        do jc=1,COL
        do ic=1,COL
          u1o%s(itb,iz,iy,ix)%u(ic,jc)=conjg(u1e%s(itb2,iz2,iy2,ix2)%u(jc,ic))
        enddo
        enddo
      enddo
    enddo
    enddo
    enddo
  case default
    call error_stop("error with |mu|>NDIM in assign_conjg_shift1_sfg_eo_wg.")
  end select

  return
end subroutine

subroutine assign_shift2_sfg_eo_wg(u2e,u1e,mu,nu)
!
! u2e(n) <= u1e(n+mu+nu)  (mu /= nu)
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u2e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e
  integer, intent(in) :: mu,nu
  integer :: ix,iy,iz,ieoxyz,itb
  integer :: ix2,iy2,iz2,itb2,ic,jc

  if (mu == nu) then
    call error_stop("error with mu == nu in assign_shift2_sfg_eo_wg.")
  endif
  if ((ABS(mu) > NDIM) .OR. (ABS(nu) > NDIM)) then
    call error_stop("error with |mu| > NDIM or |nu| > NDIM in assign_shift2_sfg_eo_wg.")
  endif

  u2e%ieo=u1e%ieo
  if ((mu < 0) .AND. (nu < 0)) then
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u2e%ieo,2)
      ix2=ix-isx(-mu)-isx(-nu)
      iy2=iy-isy(-mu)-isy(-nu)
      iz2=iz-isz(-mu)-isz(-nu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb-(1-ieoxyz)*ist(-mu)-(1-ieoxyz)*ist(-nu)
        u2e%s(itb,iz,iy,ix)=u1e%s(itb2,iz2,iy2,ix2)
      enddo
    enddo
    enddo
    enddo
  else if ((mu > 0) .AND. (nu < 0)) then
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u2e%ieo,2)
      ix2=ix+isx(mu)-isx(-nu)
      iy2=iy+isy(mu)-isy(-nu)
      iz2=iz+isz(mu)-isz(-nu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb+(  ieoxyz)*ist(mu)-(1-ieoxyz)*ist(-nu)
        u2e%s(itb,iz,iy,ix)=u1e%s(itb2,iz2,iy2,ix2)
      enddo
    enddo
    enddo
    enddo
  else if ((mu < 0) .AND. (nu > 0)) then
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u2e%ieo,2)
      ix2=ix-isx(-mu)+isx(nu)
      iy2=iy-isy(-mu)+isy(nu)
      iz2=iz-isz(-mu)+isz(nu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb-(1-ieoxyz)*ist(-mu)+(  ieoxyz)*ist(nu)
        u2e%s(itb,iz,iy,ix)=u1e%s(itb2,iz2,iy2,ix2)
      enddo
    enddo
    enddo
    enddo
  else if ((mu > 0) .AND. (nu > 0)) then
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u2e%ieo,2)
      ix2=ix+isx(mu)+isx(nu)
      iy2=iy+isy(mu)+isy(nu)
      iz2=iz+isz(mu)+isz(nu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb+(  ieoxyz)*ist(mu)+(  ieoxyz)*ist(nu)
        u2e%s(itb,iz,iy,ix)=u1e%s(itb2,iz2,iy2,ix2)
      enddo
    enddo
    enddo
    enddo
  endif

  return
end subroutine

subroutine assign_conjg_shift2_sfg_eo_wg(u2e,u1e,mu,nu)
!
! u2e(n) <= (u1e(n+mu+nu))^+  (mu /= nu)
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u2e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e
  integer, intent(in) :: mu,nu
  integer :: ix,iy,iz,ieoxyz,itb
  integer :: ix2,iy2,iz2,itb2,ic,jc

  if (mu == nu) then
    call error_stop("error with mu == nu in assign_conjg_shift2_sfg_eo_wg.")
  endif
  if ((ABS(mu) > NDIM) .OR. (ABS(nu) > NDIM)) then
    call error_stop("error with |mu| > NDIM or |nu| > NDIM in assign_conjg_shift2_sfg_eo_wg.")
  endif

  u2e%ieo=u1e%ieo
  if ((mu < 0) .AND. (nu < 0)) then
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2,ic,jc)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u2e%ieo,2)
      ix2=ix-isx(-mu)-isx(-nu)
      iy2=iy-isy(-mu)-isy(-nu)
      iz2=iz-isz(-mu)-isz(-nu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb-(1-ieoxyz)*ist(-mu)-(1-ieoxyz)*ist(-nu)
        do jc=1,COL
        do ic=1,COL
          u2e%s(itb,iz,iy,ix)%u(ic,jc)=conjg(u1e%s(itb2,iz2,iy2,ix2)%u(jc,ic))
        enddo
        enddo
      enddo
    enddo
    enddo
    enddo
  else if ((mu > 0) .AND. (nu < 0)) then
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2,ic,jc)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u2e%ieo,2)
      ix2=ix+isx(mu)-isx(-nu)
      iy2=iy+isy(mu)-isy(-nu)
      iz2=iz+isz(mu)-isz(-nu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb+(  ieoxyz)*ist(mu)-(1-ieoxyz)*ist(-nu)
        do jc=1,COL
        do ic=1,COL
          u2e%s(itb,iz,iy,ix)%u(ic,jc)=conjg(u1e%s(itb2,iz2,iy2,ix2)%u(jc,ic))
        enddo
        enddo
      enddo
    enddo
    enddo
    enddo
  else if ((mu < 0) .AND. (nu > 0)) then
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2,ic,jc)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u2e%ieo,2)
      ix2=ix-isx(-mu)+isx(nu)
      iy2=iy-isy(-mu)+isy(nu)
      iz2=iz-isz(-mu)+isz(nu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb-(1-ieoxyz)*ist(-mu)+(  ieoxyz)*ist(nu)
        do jc=1,COL
        do ic=1,COL
          u2e%s(itb,iz,iy,ix)%u(ic,jc)=conjg(u1e%s(itb2,iz2,iy2,ix2)%u(jc,ic))
        enddo
        enddo
      enddo
    enddo
    enddo
    enddo
  else if ((mu > 0) .AND. (nu > 0)) then
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ix2,iy2,iz2,itb2,ic,jc)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+u2e%ieo,2)
      ix2=ix+isx(mu)+isx(nu)
      iy2=iy+isy(mu)+isy(nu)
      iz2=iz+isz(mu)+isz(nu)
      do itb=1-ieoxyz,NTH-ieoxyz
        itb2=itb+(  ieoxyz)*ist(mu)+(  ieoxyz)*ist(nu)
        do jc=1,COL
        do ic=1,COL
          u2e%s(itb,iz,iy,ix)%u(ic,jc)=conjg(u1e%s(itb2,iz2,iy2,ix2)%u(jc,ic))
        enddo
        enddo
      enddo
    enddo
    enddo
    enddo
  endif

  return
end subroutine

subroutine assign_conjg_sfg_eo_wg(u2e,u1e)
!
!  u2e <= u1e
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u2e 
  type(sfield_gluon_eo_wg), intent(in)    :: u1e 
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc

  u2e%ieo=u1e%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+u1e%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do ic=1,COL
      do jc=1,COL
        u2e%s(itb,iz,iy,ix)%u(ic,jc)=conjg(u1e%s(itb,iz,iy,ix)%u(jc,ic))
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine assign_mult_sfg_eo_wg(u3e,u1e,u2e)
!
!  u3e <= u1e * u2e
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u3e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e,u2e
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc

#ifdef _DEBUG_
  if (u1e%ieo /= u2e%ieo) then
    call error_stop("error with u1e%ieo /= u2e%ieo in assign_mult_sfg_eo_wg.")
  endif
#endif
  u3e%ieo=u1e%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+u3e%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        u3e%s(itb,iz,iy,ix)%u(ic,jc)=u1e%s(itb,iz,iy,ix)%u(ic,1)  &
 &                                  *u2e%s(itb,iz,iy,ix)%u(1,jc)  &
 &                                  +u1e%s(itb,iz,iy,ix)%u(ic,2)  &
 &                                  *u2e%s(itb,iz,iy,ix)%u(2,jc)  &
 &                                  +u1e%s(itb,iz,iy,ix)%u(ic,3)  &
 &                                  *u2e%s(itb,iz,iy,ix)%u(3,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine assign_conjg1_mult_sfg_eo_wg(u3e,u1e,u2e)
!
!  u3e <= (u1e)^+ * u2e
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u3e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e,u2e
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc

#ifdef _DEBUG_
  if (u1e%ieo /= u2e%ieo) then
    call error_stop("error with u1e%ieo /= u2e%ieo in assign_conjg1_mult_sfg_eo_wg.")
  endif
#endif
  u3e%ieo=u1e%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+u3e%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        u3e%s(itb,iz,iy,ix)%u(ic,jc)=conjg(u1e%s(itb,iz,iy,ix)%u(1,ic)) &
 &                                  *      u2e%s(itb,iz,iy,ix)%u(1,jc)  &
 &                                  +conjg(u1e%s(itb,iz,iy,ix)%u(2,ic)) &
 &                                  *      u2e%s(itb,iz,iy,ix)%u(2,jc)  &
 &                                  +conjg(u1e%s(itb,iz,iy,ix)%u(3,ic)) &
 &                                  *      u2e%s(itb,iz,iy,ix)%u(3,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine assign_conjg2_mult_sfg_eo_wg(u3e,u1e,u2e)
!
!  u3e <= u1e * (u2e)^+
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u3e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e,u2e
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc

#ifdef _DEBUG_
  if (u1e%ieo /= u2e%ieo) then
    call error_stop("error with u1e%ieo /= u2e%ieo in assign_conjg2_mult_sfg_eo_wg.")
  endif
#endif
  u3e%ieo=u1e%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+u3e%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        u3e%s(itb,iz,iy,ix)%u(ic,jc)=      u1e%s(itb,iz,iy,ix)%u(ic,1)  &
 &                                  *conjg(u2e%s(itb,iz,iy,ix)%u(jc,1)) &
 &                                  +      u1e%s(itb,iz,iy,ix)%u(ic,2)  &
 &                                  *conjg(u2e%s(itb,iz,iy,ix)%u(jc,2)) &
 &                                  +      u1e%s(itb,iz,iy,ix)%u(ic,3)  &
 &                                  *conjg(u2e%s(itb,iz,iy,ix)%u(jc,3))
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine assign_rmult_sfg_eo_wg(u2e,u1e,rcoef)
!
!  u2e <= u1e * rcoef
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u2e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e
  real(DP), intent(in) :: rcoef
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc

  u2e%ieo=u1e%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+u1e%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        u2e%s(itb,iz,iy,ix)%u(ic,jc)=rcoef*u1e%s(itb,iz,iy,ix)%u(ic,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

!*************************************************************************
subroutine clear_sfg_eo_wg(ue)
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: ue
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ue%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        ue%s(itb,iz,iy,ix)%u(ic,jc)=Z0
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine clear_sfg_eo_wog(ue)
  implicit none
  type(sfield_gluon_eo_wog), intent(inout) :: ue
  integer :: ix,iy,iz,itb

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itb)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    do itb=1,NTH
      ue%s(itb,iz,iy,ix)%u(:,:)=Z0
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine clear_vfg_eo_wg(ue)
  implicit none
  type(vfield_gluon_eo_wg), intent(inout) :: ue
  integer :: mu
  do mu=1,NDIM
    call clear(ue%mu(mu))
  enddo
  return
end subroutine

subroutine clear_vfg_eo_wog(ue)
  implicit none
  type(vfield_gluon_eo_wog), intent(inout) :: ue
  integer :: mu
  do mu=1,NDIM
    call clear(ue%mu(mu))
  enddo
  return
end subroutine

subroutine clear_vfg_wg(u)
  implicit none
  type(vfield_gluon_wg), intent(inout) :: u
  call clear(u%eo(0))
  call clear(u%eo(1))
  return
end subroutine

subroutine clear_vfg_wog(u)
  implicit none
  type(vfield_gluon_wog), intent(inout) :: u
  call clear(u%eo(0))
  call clear(u%eo(1))
  return
end subroutine

subroutine clear_tfg_eo_wg(ue)
  implicit none
  type(tfield_gluon_eo_wg), intent(inout) :: ue
  integer :: mu
  do mu=1,NDIM*(NDIM-1)/2
    call clear(ue%munu(mu))
  enddo
  return
end subroutine

subroutine clear_tfg_wg(u)
  implicit none
  type(tfield_gluon_wg), intent(inout) :: u
  call clear(u%eo(0))
  call clear(u%eo(1))
  return
end subroutine


!*************************************************************************
subroutine assign_add_sfg_eo_wg(u3e,u1e,u2e)
!
! u3e <= u1e + u2e
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u3e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e,u2e
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc

#ifdef _DEBUG_
  if (u1e%ieo /= u2e%ieo) then
    call error_stop("error with u1e%ieo /= u2e%ieo in assign_add_sfg_eo_wg.")
  endif
#endif
  u3e%ieo=u1e%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+u1e%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        u3e%s(itb,iz,iy,ix)%u(ic,jc)=u1e%s(itb,iz,iy,ix)%u(ic,jc)  &
 &                                  +u2e%s(itb,iz,iy,ix)%u(ic,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine accum_add_sfg_eo_wg(u2e,u1e)
!
! u2e <= u2e + u1e
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u2e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc

#ifdef _DEBUG_
  if (u1e%ieo /= u2e%ieo) then
    call error_stop("error with u1e%ieo /= u2e%ieo in accum_add_sfg_eo_wg.")
  endif
#endif
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+u1e%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        u2e%s(itb,iz,iy,ix)%u(ic,jc)=u2e%s(itb,iz,iy,ix)%u(ic,jc)  &
 &                                  +u1e%s(itb,iz,iy,ix)%u(ic,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine assign_sub_sfg_eo_wg(u3e,u1e,u2e)
!
! u3e = u1e - u2e
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u3e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e,u2e
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc

#ifdef _DEBUG_
  if (u1e%ieo /= u2e%ieo) then
    call error_stop("error with u1e%ieo /= u2e%ieo in assign_sub_sfg_eo_wg.")
  endif
#endif
  u3e%ieo=u1e%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+u1e%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        u3e%s(itb,iz,iy,ix)%u(ic,jc)=u1e%s(itb,iz,iy,ix)%u(ic,jc)  &
 &                                  -u2e%s(itb,iz,iy,ix)%u(ic,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine accum_sub_sfg_eo_wg(u2e,u1e)
!
! u2e <= u2e + u1e
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: u2e
  type(sfield_gluon_eo_wg), intent(in)    :: u1e
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc

#ifdef _DEBUG_
  if (u1e%ieo /= u2e%ieo) then
    call error_stop("error with u1e%ieo /= u2e%ieo in accum_sub_sfg_eo_wg.")
  endif
#endif
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+u1e%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        u2e%s(itb,iz,iy,ix)%u(ic,jc)=u2e%s(itb,iz,iy,ix)%u(ic,jc)  &
 &                                  -u1e%s(itb,iz,iy,ix)%u(ic,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

function trace_sfg_eo_wg(ue) result(my_trace)
!
! return trace of gauge fields (even/odd sites only)
!
  use comlib
  implicit none
  type(sfield_gluon_eo_wg), intent(in) :: ue
  complex(DP) :: my_trace
  integer :: ix,iy,iz,ieoxyz,itb,ic

  my_trace = Z0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb) REDUCTION(+:my_trace)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ue%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      do ic=1,COL
        my_trace = my_trace+ue%s(itb,iz,iy,ix)%u(ic,ic)
      enddo
    enddo
  enddo
  enddo
  enddo
#ifndef _singlePU
  call comlib_sumcast(my_trace)
#endif

  return
end function

subroutine new_fields_gauge(this)
!
! Initialize gauge field parallel environment (communication)
!
  implicit none
  type(fields_gauge), intent(inout) :: this  ! contain initialization status
  integer :: imu(NDIM*(NDIM-1)/2),mu
  type(lattice_world) :: lattice

  if (.not. is_initialized(lattice)) then
    write(*,'("Lattice is not initialized!")')
    stop
  endif

  if (m_is_initialized) then
    write(*,'("Gauge Fields are already initialized!")')
    return
  endif

  call new(copy_sfg_time)

  imu(1)=1
  imu(2)=2
  imu(3)=3
  imu(4)=1
  imu(5)=1
  imu(6)=2

  if (NPU > 1) then
    do mu=1,(NDIM-1)*2
#ifndef _singlePU
      gbuffup(:,1,mu) = Z0
      gbuffup(:,2,mu) = Z0
      gbuffdn(:,1,mu) = Z0
      gbuffdn(:,2,mu) = Z0
!******** mu-upward-direction comlib setup (gauge)
      call comlib_make2(idgsendup(mu),nodeidup(imu(mu)),   &
                        gbuffup(:,1,mu),gbuffup(:,2,mu),   &
                        GBUFF_SIZE(mu))
      write(idgsendup(mu)%name,'("COPY_U_UP:",I2)')mu

!******** mu-downward-direction comlib setup (gauge)
      call comlib_make2(idgsenddn(mu),nodeiddn(imu(mu)),   &
                        gbuffdn(:,1,mu),gbuffdn(:,2,mu),   &
                        GBUFF_SIZE(mu))
      write(idgsenddn(mu)%name,'("COPY_U_DN:",I2)')mu
#endif
    enddo
  endif

  m_is_initialized = .TRUE.

  return
end subroutine

function abs2_sfg_eo_wog(pe) result(my_abs2)
!
! return squared norm of gauge field (momentum)
!
  implicit none
  type(sfield_gluon_eo_wog), intent(in) :: pe
  real(DP) :: my_abs2
  integer :: ix,iy,iz,itb,ic,jc

  my_abs2 = 0.0_DP
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itb,ic,jc) REDUCTION(+:my_abs2)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    do itb=1,NTH
      do jc=1,COL
      do ic=1,COL
        my_abs2 = my_abs2+ real(pe%s(itb,iz,iy,ix)%u(ic,jc))**2  &
 &                       +aimag(pe%s(itb,iz,iy,ix)%u(ic,jc))**2
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo
#ifndef _singlePU
  call comlib_sumcast(my_abs2)
#endif

  return
end function

subroutine assign_mult_sfg_eo_wg_wog(pe,u1e,u2e)
!
!  pe <= u1e * u2e
!
  implicit none
  type(sfield_gluon_eo_wog), intent(inout) :: pe
  type(sfield_gluon_eo_wg),  intent(in)    :: u1e,u2e
  integer :: ix,iy,iz,ieoxyz,itb,ic,jc,itb0

#ifdef _DEBUG_
  if (u1e%ieo /= u2e%ieo) then
    call error_stop("error with u1e%ieo /= u2e%ieo in assign_mult_sfg_eo_wog.")
  endif
#endif
  pe%ieo=u1e%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,itb0,ic,jc)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+pe%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      itb0 = itb+ieoxyz
      do jc=1,COL
      do ic=1,COL
        pe%s(itb0,iz,iy,ix)%u(ic,jc)=u1e%s(itb,iz,iy,ix)%u(ic,1)  &
 &                                  *u2e%s(itb,iz,iy,ix)%u(1,jc)  &
 &                                  +u1e%s(itb,iz,iy,ix)%u(ic,2)  &
 &                                  *u2e%s(itb,iz,iy,ix)%u(2,jc)  &
 &                                  +u1e%s(itb,iz,iy,ix)%u(ic,3)  &
 &                                  *u2e%s(itb,iz,iy,ix)%u(3,jc)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine assign_mult_vfg_wg_wog(p,u,BB)
!
!  pe <= u1e * u2e
!
  implicit none
  type(vfield_gluon_wog), intent(inout) :: p
  type(vfield_gluon_wg),  intent(in)    :: u,BB
  integer :: mu

  do mu=1,NDIM
    call assign_mult(p%eo(0)%mu(mu),u%eo(0)%mu(mu),BB%eo(0)%mu(mu))
    call assign_mult(p%eo(1)%mu(mu),u%eo(1)%mu(mu),BB%eo(1)%mu(mu))
  enddo

  return
end subroutine

subroutine accum_add_sfg_eo_wg_eo_wog(pe,ue)
!
! pe <= pe + ue
!
  implicit none
  type(sfield_gluon_eo_wog), intent(inout) :: pe
  type(sfield_gluon_eo_wg),  intent(in)    :: ue
  integer :: ix,iy,iz,ieoxyz,itb,itb0

#ifdef _DEBUG_
  if (pe%ieo /= ue%ieo) then
    call error_stop("error with pe%ieo /= ue%ieo in accum_add_sfg_eo_wg_eo_wog.")
  endif
#endif
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,itb0)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ue%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      itb0 = itb+ieoxyz
      pe%s(itb0,iz,iy,ix)%u(:,:)=pe%s(itb0,iz,iy,ix)%u(:,:)  &
 &                              +ue%s(itb ,iz,iy,ix)%u(:,:)
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine traceless_ahermite_sfg_eo_wog(pe)
!
!  pe <= (pe - pe^+) - tr[pe - pe^+]/3
!
  implicit none
  type(sfield_gluon_eo_wog), intent(inout) :: pe
  complex(DP) :: trace,ah(COL,COL)
  integer :: ix,iy,iz,itb,ic,jc

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itb,ic,jc,trace,ah)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    do itb=1,NTH
      trace = ( pe%s(itb,iz,iy,ix)%u(1,1)       &
 &             +pe%s(itb,iz,iy,ix)%u(2,2)       &
 &             +pe%s(itb,iz,iy,ix)%u(3,3))/COL

      do ic=1,COL
        pe%s(itb,iz,iy,ix)%u(ic,ic)=pe%s(itb,iz,iy,ix)%u(ic,ic) - trace
      enddo

      do jc=1,COL
      do ic=1,COL
        ah(ic,jc) = pe%s(itb,iz,iy,ix)%u(ic,jc)  &
 &           -conjg(pe%s(itb,iz,iy,ix)%u(jc,ic))
      enddo
      enddo

      do jc=1,COL
      do ic=1,COL
        pe%s(itb,iz,iy,ix)%u(ic,jc) = ah(ic,jc)
      enddo
      enddo

    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine traceless_ahermite_vfg_wog(p)
!
!  p <= (p - p^+) - tr[p - p^+]/3
!
  implicit none
  type(vfield_gluon_wog), intent(inout) :: p
  integer :: mu
  do mu=1,NDIM
    call traceless_antihermite(p%eo(0)%mu(mu))
    call traceless_antihermite(p%eo(1)%mu(mu))
  enddo
  return
end subroutine


subroutine accum_add_cmult_sfg_eo_wog(p2e,p1e,coef)
!
!  p2e <= p2e + p1e * coef
!
  implicit none
  type(sfield_gluon_eo_wog), intent(inout) :: p2e
  type(sfield_gluon_eo_wog), intent(in)    :: p1e
  complex(DP), intent(in) :: coef
  integer :: ix,iy,iz,itb

  p2e%ieo=p1e%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itb)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    do itb=1,NTH
      p2e%s(itb,iz,iy,ix)%u(:,:)=p2e%s(itb,iz,iy,ix)%u(:,:)  &
 &                         +coef*p1e%s(itb,iz,iy,ix)%u(:,:)
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine set_identity_u(u)
!
! Initialize gauge field to unit matrix (cold config)
!
  implicit none
  type(vfield_gluon_wg), intent(inout) :: u ! gauge field

  integer :: ic,jc,mu,ix,iy,iz,itb

!************************************************************
!* Set initial gauge link
!************************************************************
  call new(u)
  do mu = 1,NDIM
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itb,ic,jc)
  do ix = 0,NX1
  do iy = 0,NY1
  do iz = 0,NZ1
  do itb = 0,NTH
    do jc = 1,COL
    do ic = 1,COL
      u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc) = Z0
      u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc) = Z0
    enddo
    enddo
    do ic = 1,COL
      u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(ic,ic) = Z1
      u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(ic,ic) = Z1
    enddo
  enddo
  enddo
  enddo
  enddo
  enddo

  call copy_boundary(u)
#ifdef _SF
  call set_sf_boundary(u)
#endif

  return
end subroutine

subroutine set_gaussian_noise_p(p)
!
! \set Gaussian noise on canonical momentum (su(3) Lie algebra) of gauge field (SU(3) Lie group)
!
  implicit none
  type(vfield_gluon_wog), intent(inout) :: p   !  gauge momentum field

  integer :: COLADJ
  parameter (COLADJ=COL**2-1)
  real(DP) :: He(COLADJ,0:NTH),Ho(COLADJ,0:NTH)
  complex(DP) :: SU(COL,COL,COLADJ)
  real(DP) :: pi2
  real(DP) :: th(COLADJ*NTH),yr(COLADJ*NTH),sth(COLADJ*NTH)
  integer :: ix,iy,iz,itb,ieoxyz,ic,jc,mu,icad,itb0
  integer :: ictb,ir
#ifdef _PARALLELCHECK
  integer :: itx,ity,itz,itt,ipx,ipy,ipz,ieo
  real(DP) :: HH(COLADJ)
  real(DP) :: pth(COLADJ),pyr(COLADJ)
#endif

  pi2 = PI*2
  call new(p)

  do icad = 1,COLADJ
    do jc = 1,COL
    do ic = 1,COL
      SU(ic,jc,icad) = z0
    enddo
    enddo
  enddo

!****************************
!* set su(3) generator 
!* (times sqrt(2))
!****************************
  SU(1,2,1) =   z1/sqrt(2.0_DP)
  SU(2,1,1) =   z1/sqrt(2.0_DP)
  SU(1,2,2) =  -zi/sqrt(2.0_DP)
  SU(2,1,2) =   zi/sqrt(2.0_DP)
  SU(1,1,3) =   z1/sqrt(2.0_DP)
  SU(2,2,3) =  -z1/sqrt(2.0_DP)
  SU(1,3,4) =   z1/sqrt(2.0_DP)
  SU(3,1,4) =   z1/sqrt(2.0_DP)
  SU(1,3,5) =  -zi/sqrt(2.0_DP)
  SU(3,1,5) =   zi/sqrt(2.0_DP)
  SU(2,3,6) =   z1/sqrt(2.0_DP)
  SU(3,2,6) =   z1/sqrt(2.0_DP)
  SU(2,3,7) =  -zi/sqrt(2.0_DP)
  SU(3,2,7) =   zi/sqrt(2.0_DP)
  SU(1,1,8) =   z1/sqrt(6.0_DP)
  SU(2,2,8) =   z1/sqrt(6.0_DP)
  SU(3,3,8) =-2*z1/sqrt(6.0_DP)

#ifdef _PARALLELCHECK
!***********************************
!* programmed independent on 
!* parallelization size
!***********************************
  do mu=1,NDIM
  do itx=1,NTX
  do ity=1,NTY
  do itz=1,NTZ
  do itt=1,NTT
    ieo = mod(itx+ity+itz+itt,2)
    ipx = (itx-1)/NX
    ipy = (ity-1)/NY
    ipz = (itz-1)/NZ
    ix  = mod(itx-1,NX)+1
    iy  = mod(ity-1,NY)+1
    iz  = mod(itz-1,NZ)+1
    itb = itt/2

    call get(g_rand,pyr)
    call get(g_rand,pth)

    if ( (ipx == ipsite(1)) .and. &
 &       (ipy == ipsite(2)) .and. &
 &       (ipz == ipsite(3)) ) then

      ieoxyz = mod(ipeo+iz+iy+ix+ieo,2)
      itb0 = itb + ieoxyz

      do icad=1,COLADJ
        pth(icad) = pth(icad)*pi2
        pyr(icad) = SQRT(-2.0_DP*LOG(pyr(icad)))
        pth(icad) =              SIN(pth(icad))
         HH(icad) = pyr(icad)*pth(icad)
      enddo
      do jc = 1,COL
      do ic = 1,COL
        p%eo(ieo)%mu(mu)%s(itb0,iz,iy,ix)%u(ic,jc)=HH(1)*SU(ic,jc,1)  &
 &                                                +HH(2)*SU(ic,jc,2)  &
 &                                                +HH(3)*SU(ic,jc,3)  &
 &                                                +HH(4)*SU(ic,jc,4)  &
 &                                                +HH(5)*SU(ic,jc,5)  &
 &                                                +HH(6)*SU(ic,jc,6)  &
 &                                                +HH(7)*SU(ic,jc,7)  &
 &                                                +HH(8)*SU(ic,jc,8)
      enddo
      enddo
    endif
  enddo
  enddo
  enddo
  enddo
  enddo
#else

!********************************
!* set initial momentum
!* with parallel rundom number
!* generator (M series).
!*
!* Outer (ix,iy,iz) loop is 
!* divided by NPULOCAL to achieve
!* better efficiency for element
!* parallelization.
!********************************
  do mu = 1,NDIM
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz,2)

!********************
!* even sites
!********************
      call get(g_rand,yr)
      call get(g_rand,th)
      do ictb=1,COLADJ*NTH
         yr(ictb) = SQRT(-2.0_DP*LOG(yr(ictb)))
        sth(ictb) = SIN(th(ictb)*pi2)
      enddo
      ir=1
      do itb=1-ieoxyz,NTH-ieoxyz
        do ic = 1,COLADJ
          He(ic,itb) = yr(ir)*sth(ir)
          ir=ir+1
        enddo
      enddo

!********************
!* odd sites
!********************
      call get(g_rand,yr)
      call get(g_rand,th)
      do ictb=1,COLADJ*NTH
         yr(ictb) = SQRT(-2.0_DP*LOG(yr(ictb)))
        sth(ictb) = SIN(th(ictb)*pi2)
      enddo
      ir=1
      do itb=ieoxyz,NTH+ieoxyz-1
        do ic = 1,COLADJ
          Ho(ic,itb) = yr(ir)*sth(ir)
          ir=ir+1
        enddo
      enddo

!********************
!* even sites
!********************
      do itb=1-ieoxyz,NTH-ieoxyz
        itb0 = itb + ieoxyz
        do jc = 1,COL
        do ic = 1,COL
          p%eo(0)%mu(mu)%s(itb0,iz,iy,ix)%u(ic,jc)=He(1,itb)*SU(ic,jc,1)  &
 &                                                +He(2,itb)*SU(ic,jc,2)  &
 &                                                +He(3,itb)*SU(ic,jc,3)  &
 &                                                +He(4,itb)*SU(ic,jc,4)  &
 &                                                +He(5,itb)*SU(ic,jc,5)  &
 &                                                +He(6,itb)*SU(ic,jc,6)  &
 &                                                +He(7,itb)*SU(ic,jc,7)  &
 &                                                +He(8,itb)*SU(ic,jc,8)
        enddo
        enddo
      enddo

!********************
!* odd sites
!********************
      do itb=ieoxyz,NTH+ieoxyz-1
        itb0 = 1+itb - ieoxyz
        do jc = 1,COL
        do ic = 1,COL
          p%eo(1)%mu(mu)%s(itb0,iz,iy,ix)%u(ic,jc)=Ho(1,itb)*SU(ic,jc,1)  &
 &                                                +Ho(2,itb)*SU(ic,jc,2)  &
 &                                                +Ho(3,itb)*SU(ic,jc,3)  &
 &                                                +Ho(4,itb)*SU(ic,jc,4)  &
 &                                                +Ho(5,itb)*SU(ic,jc,5)  &
 &                                                +Ho(6,itb)*SU(ic,jc,6)  &
 &                                                +Ho(7,itb)*SU(ic,jc,7)  &
 &                                                +Ho(8,itb)*SU(ic,jc,8)
        enddo
        enddo
      enddo
  enddo
  enddo
  enddo ! end of do ix,iy,iz
  enddo ! end of do mu
#endif

#ifdef _SF
  call set_sf_boundary(p)
#endif

  return
end subroutine

subroutine plaquette_u(u,plq)
!
! Calculate space-time-direction averaged plaquette
!
  use comlib
  implicit none
  type(vfield_gluon_wg), intent(in)  :: u    ! gauge field
  real(DP),              intent(out) :: plq  ! averaged plaquett
  real(DP) :: plqe,plqo

  call plaquette_u_eo(u%eo(0),u%eo(1),plqe)
  call plaquette_u_eo(u%eo(1),u%eo(0),plqo)

  plq=plqe+plqo

  return
end subroutine

subroutine plaquette_u_eo(ue,uo,plqe)
!
! Calculate space-time-direction averaged plaquette on even/odd sites
!
  implicit none
  type(vfield_gluon_eo_wg), intent(in)  :: ue   ! even/odd site gauge field
  type(vfield_gluon_eo_wg), intent(in)  :: uo   ! odd/even site gauge field
  real(DP),                 intent(out) :: plqe ! averaged plaquett on even/odd site

  type(sfield_gluon_eo_wg), allocatable :: u12e,u14e,we
  integer :: mu,nu
  real(DP) :: rtmp0

  allocate(u12e,u14e,we)
  rtmp0 = 0.0_DP
  do mu = 1,NDIM
  do nu = 1,NDIM
    if (mu == nu) cycle

!*********************************
!*  nu
!*
!*           3
!*           |
!*  u14e     ^  we
!*           |
!*   n--->---2  mu
!*
!*********************************
    call assign_shift(we,  uo%mu(nu),mu)
    call  assign_mult(u12e,ue%mu(mu),we)

!*********************************
!*  nu
!*      we
!*   4--->---3
!*   |        
!*   ^   u14e
!*   |        
!*   n       2  mu
!*
!*********************************
    call assign_shift(we,  uo%mu(mu),nu)
    call  assign_mult(u14e,ue%mu(nu),we)

!*********************************
!*  nu
!*          
!*   4---<---3
!*   |       |
!*   v       ^  = u12e*(u14e)^+
!*   |       |
!*   n--->---2  mu
!*
!*********************************
    call assign_conjg2_mult(we,u12e,u14e)
    rtmp0=rtmp0+real(trace(we))

  enddo
  enddo

  deallocate(u12e,u14e,we)

  plqe=rtmp0/(NTX*NTY*NTZ*NTT)/(NDIM-1)/NDIM/COL

  return
end subroutine

subroutine update_p_wog(p,dt,F)
!
! update gauge momentum field using force F
!
!  p_mu(n) = p_mu(n) - i dt F_mu(n)
!
!  Force <= U *(D Hamil / D U) (<= fix me, or check!)
!
  implicit none
  type(vfield_gluon_wog), intent(inout) :: p    ! gauge momentum field
  real(DP),               intent(in)    :: dt   ! MD time step
  type(vfield_gluon_wog), intent(in)    :: F    ! gauge force field

  call update_p_eo_wog(p%eo(0),dt,F%eo(0))
  call update_p_eo_wog(p%eo(1),dt,F%eo(1))

  return
end subroutine

subroutine update_p_eo_wog(pe,dt,Fe)
!
! update gaugem mementum field on even/odd sites using force field.
!
  implicit none
  type(vfield_gluon_eo_wog), intent(inout) :: pe  ! gauge momentum field
  real(DP),                  intent(in)    :: dt  ! MD step size
  type(vfield_gluon_eo_wog), intent(in)    :: Fe  ! gauge force field

  integer :: mu
  complex(DP) :: zidt

#ifdef _DEBUG_
  if ( pe%ieo /= Fe%ieo ) then
    call error_stop("error with pe%ieo /= Fe%ieo in update_p_eo_wog.")
  endif
#endif
  
  zidt=-zi*dt
  do mu=1,NDIM
    call accum_add_mult(pe%mu(mu),Fe%mu(mu),zidt)
  enddo ! end of do mu

#ifdef _SF
  call set_sf_boundary(pe)
#endif

  return
end subroutine

subroutine update_u(u,dt,p)
!
! Update gauge field using canonical momentum :
!
!    u_mu(dt)=exp(i p_mu(dt')dt) u_mu(0)
!
  implicit none
  type(vfield_gluon_wg),  intent(inout) :: u   ! gauge link field
  real(DP),               intent(in)    :: dt  ! MD step size
  type(vfield_gluon_wog), intent(in)    :: p   ! gauge momentum field

  call update_u_eo(u%eo(0),dt,p%eo(0))
  call update_u_eo(u%eo(1),dt,p%eo(1))

  return
end subroutine

subroutine update_u_eo(ue,dt,pe)
!
! Update gauge field using canonical momentum on eve/odd sites
!
!    u_mu(dt)=exp(i p_mu(dt')dt) u_mu(0)
!
  implicit none
  type(vfield_gluon_eo_wg),  intent(inout) :: ue ! gauge link field on even/odd sites
  real(DP),                  intent(in)    :: dt ! MD step size
  type(vfield_gluon_eo_wog), intent(in)    :: pe ! gauge momentum field on even/odd sites

  real(DP) :: rp(COL),rq(COL),fnorm,tb,tc
  real(DP) :: sqrp(COL),sqzp(COL)
  complex(DP) :: zp(COL),zq(COL),zcoef,ecoef(COL),aa(COL),ep(COL,COL),epu(COL,COL)
  integer :: iexp,istep,nstep
  integer :: ix,iy,iz,itb,ieoxyz,itb0,mu,ic,jc

#ifdef _DEBUG_
  if ( ue%ieo /= pe%ieo) then
    call error_stop("error with ue%ieo /= pe%ieo in update_u_eo.")
  endif
#endif

  do mu=1,NDIM
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,ieoxyz,itb,itb0,ic,jc,iexp,istep,nstep, &
!$OMP         rp,zp,rq,zq,sqrp,sqzp,tb,tc,fnorm,ep,epu,ecoef,aa,zcoef)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+pe%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      itb0 = itb+ieoxyz

      !
      ! Hermitian matrix parametrization
      !
      !  rx(1)    zx(3)^*  zx(2)
      !  zx(3)    rx(2)    zx(1)^*
      !  zx(2)^*  zx(1)    rx(3)
      !
      rp(1) = dt*pe%mu(mu)%s(itb0,iz,iy,ix)%u(1,1)
      rp(2) = dt*pe%mu(mu)%s(itb0,iz,iy,ix)%u(2,2)
      rp(3) = -(rp(1)+rp(2))
      zp(1) = dt*pe%mu(mu)%s(itb0,iz,iy,ix)%u(3,2)
      zp(2) = dt*pe%mu(mu)%s(itb0,iz,iy,ix)%u(1,3)
      zp(3) = dt*pe%mu(mu)%s(itb0,iz,iy,ix)%u(2,1)

      !
      ! compute Frobenius norm of P
      !
      fnorm = ( rp(1)**2 + rp(2)**2 + rp(1)*rp(2) &
 &            + real(zp(1))**2 + aimag(zp(1))**2  &
 &            + real(zp(2))**2 + aimag(zp(2))**2  &
 &            + real(zp(3))**2 + aimag(zp(3))**2 )*2.0_DP
      fnorm = SQRT(fnorm)

      !
      ! |dt P| = 0 then skip U = exp(i dt P) U
      !
      if (fnorm == ZERO) cycle

      !
      ! compute step number for scaling-squaring method
      !
      nstep = MAX(CEILING(LOG(2*fnorm)/LOG(2.0_DP)),0)

      !
      ! rescale P as  P <= P/2^nstep
      !
      fnorm = 1.0_DP/(2**nstep)
      rp(1) = rp(1)*fnorm
      rp(2) = rp(2)*fnorm
      rp(3) = -(rp(1)+rp(2))
      zp(1) = zp(1)*fnorm
      zp(2) = zp(2)*fnorm
      zp(3) = zp(3)*fnorm

      sqrp(1) = rp(1)**2
      sqrp(2) = rp(2)**2
      sqrp(3) = rp(3)**2
      sqzp(1) = REAL(zp(1))**2 + AIMAG(zp(1))**2
      sqzp(2) = REAL(zp(2))**2 + AIMAG(zp(2))**2
      sqzp(3) = REAL(zp(3))**2 + AIMAG(zp(3))**2

      !
      ! Squared P
      !
      ! Q = P^2
      !
      rq(1) = sqrp(1) + sqzp(3) + sqzp(2)
      rq(2) = sqzp(3) + sqrp(2) + sqzp(1)
      rq(3) = sqzp(2) + sqzp(1) + sqrp(3)
      zq(1) = zp(1)*(rp(2)+rp(3)) + CONJG(zp(2)*zp(3))
      zq(2) = zp(2)*(rp(3)+rp(1)) + CONJG(zp(3)*zp(1))
      zq(3) = zp(3)*(rp(1)+rp(2)) + CONJG(zp(1)*zp(2))

      !
      ! tb = Tr[P^2]/2 = Tr[Q]/2
      !
      tb = (rq(1) + rq(2) + rq(3))/2.0_DP

      !
      ! tc = Tr[P^3]/3
      !
      sqrp(1) = sqrp(1)*2.0_DP/3.0_DP
      sqrp(2) = sqrp(2)*2.0_DP/3.0_DP
      sqrp(3) = sqrp(3)*2.0_DP/3.0_DP
      tc = rp(1)*(rq(1)-sqrp(1))  &
 &       + rp(2)*(rq(2)-sqrp(2))  &
 &       + rp(3)*(rq(3)-sqrp(3))  &
 &       + REAL(zp(1)*zp(2)*zp(3))*2.0_DP

      !
      ! Compute matrix exponential coefficinets with Cayley-Hamilton method
      !
      ecoef(1) = Z1
      ecoef(2) = Z0
      ecoef(3) = Z0
      do iexp = EXP_TAYLOR_ORDER,1,-1
        zcoef = ZI/iexp
        aa(1) = tc*ecoef(3)
        aa(2) = tb*ecoef(3)+ecoef(1)
        aa(3) =             ecoef(2)
        ecoef(1) = aa(1)*zcoef + Z1
        ecoef(2) = aa(2)*zcoef
        ecoef(3) = aa(3)*zcoef
      enddo

      !
      ! first two columns are computed
      !
      ep(1,1) = ecoef(3)*      rq(1)  + ecoef(2)*      rp(1) + ecoef(1)
      ep(2,1) = ecoef(3)*      zq(3)  + ecoef(2)*      zp(3)
      ep(3,1) = ecoef(3)*CONJG(zq(2)) + ecoef(2)*CONJG(zp(2))

      ep(1,2) = ecoef(3)*CONJG(zq(3)) + ecoef(2)*CONJG(zp(3))
      ep(2,2) = ecoef(3)*      rq(2)  + ecoef(2)*      rp(2) + ecoef(1)
      ep(3,2) = ecoef(3)*      zq(1)  + ecoef(2)*      zp(1)
      call reunit(ep)

      !
      ! Squaring step for scaling and squaring method
      ! 
      do istep=1,nstep
        ep(:,1:2) = MATMUL(ep(:,:),ep(:,1:2))
        call reunit(ep)
      enddo

      do jc=1,COL-1
      do ic=1,COL
        epu(ic,jc) = ep(ic,1)*ue%mu(mu)%s(itb,iz,iy,ix)%u(1,jc)  &
 &                 + ep(ic,2)*ue%mu(mu)%s(itb,iz,iy,ix)%u(2,jc)  &
 &                 + ep(ic,3)*ue%mu(mu)%s(itb,iz,iy,ix)%u(3,jc)
      enddo
      enddo

      call reunit(epu)

      do jc=1,COL
      do ic=1,COL
        ue%mu(mu)%s(itb,iz,iy,ix)%u(ic,jc) = epu(ic,jc)
      enddo
      enddo

    enddo
  enddo
  enddo
  enddo

  call copy_boundary(ue%mu(mu))  ! ghost site boundary exchange.

  enddo ! end of do mu

#ifdef _SF
  call set_sf_boundary(ue)
#endif

  return

contains

  subroutine reunit(u)
  !
  ! reunitalization for SU(3) matrix using
  ! GS reorthonormalization for the first two columns
  ! and SU(3) reconstruction for the third column
  !
    implicit none
    complex(DP), intent(inout) :: u(COL,COL)
    complex(DP) :: ctmp
    real(DP) :: fnorm

    fnorm = REAL(u(1,1))**2 + AIMAG(u(1,1))**2 &
 &        + REAL(u(2,1))**2 + AIMAG(u(2,1))**2 &
 &        + REAL(u(3,1))**2 + AIMAG(u(3,1))**2
    fnorm = 1.0_DP/SQRT(fnorm)
    u(1,1) = u(1,1)*fnorm
    u(2,1) = u(2,1)*fnorm
    u(3,1) = u(3,1)*fnorm
    ctmp = CONJG(u(1,1)) * u(1,2)  &
 &       + CONJG(u(2,1)) * u(2,2)  &
 &       + CONJG(u(3,1)) * u(3,2)
    u(1,2) = u(1,2) - u(1,1)*ctmp
    u(2,2) = u(2,2) - u(2,1)*ctmp
    u(3,2) = u(3,2) - u(3,1)*ctmp
    fnorm = REAL(u(1,2))**2 + AIMAG(u(1,2))**2 &
 &        + REAL(u(2,2))**2 + AIMAG(u(2,2))**2 &
 &        + REAL(u(3,2))**2 + AIMAG(u(3,2))**2
    fnorm = 1.0_DP/SQRT(fnorm)
    u(1,2) = u(1,2)*fnorm
    u(2,2) = u(2,2)*fnorm
    u(3,2) = u(3,2)*fnorm
    !
    ! third column is reconstructed from the first two columns
    ! using SU(3) property
    !
    u(1,3) = CONJG( u(2,1)*u(3,2) - u(3,1)*u(2,2) )
    u(2,3) = CONJG( u(3,1)*u(1,2) - u(1,1)*u(3,2) )
    u(3,3) = CONJG( u(1,1)*u(2,2) - u(2,1)*u(1,2) )
    return
  end subroutine
end subroutine

function get_force_norm_wog(F) result(norm)
!
! return averaged force norm
!
!  norm = SQRT(abs2(F)/(NDIM*SPACETIMEVOLUME))
!
  implicit none
  type(vfield_gluon_wog), intent(in) :: F  ! MD force
  real(DP) :: norm
  integer :: mu

  norm = 0.0_DP
  do mu=1,NDIM
    norm = norm + abs2(F%eo(0)%mu(mu))
    norm = norm + abs2(F%eo(1)%mu(mu))
  enddo
  norm = sqrt(norm/(NDIM*NTT*NTZ*NTY*NTX))

  return
end function

subroutine make_two_links(u,wl34,wl98)
!
! Prepare 2-links field (tensor) for force contribution
! 
!     4---<---3     8       n
!             |     |
!       wl34  ^     ^  wl98
!             |     |
!     n       2     9---<---6    for mu < nu
! 
! For mu < nu cases, use wl34^+, wl98^+.
! 
  implicit none
  type(vfield_gluon_wg), intent(in)    :: u     ! gauge field
  type(tfield_gluon_wg), intent(inout) :: wl34  ! even/odd site upper right 2-links
  type(tfield_gluon_wg), intent(inout) :: wl98  ! even/odd site lower left  2-links

  call make_two_links_eo(u%eo(0),u%eo(1),wl34%eo(0),wl98%eo(0))
  call make_two_links_eo(u%eo(1),u%eo(0),wl34%eo(1),wl98%eo(1))

  return
end subroutine

subroutine make_two_links_eo(ue,uo,wl34e,wl98e)
!
! Prepare 2-links field (tensor) for force contribution, even/odd sites only
! 
  implicit none
  type(vfield_gluon_eo_wg), intent(in)    :: ue     ! even/odd site gauge field
  type(vfield_gluon_eo_wg), intent(in)    :: uo     ! odd/even site gauge field
  type(tfield_gluon_eo_wg), intent(inout) :: wl34e  ! even/odd site upper right 2-links
  type(tfield_gluon_eo_wg), intent(inout) :: wl98e  ! even/odd site lower left 2-links

  type(sfield_gluon_eo_wg), allocatable :: w1e,w2e
  integer :: mu,nu,ipl

  allocate(w1e,w2e)
  call new(wl34e,ue%ieo)
  call new(wl98e,ue%ieo)

  do mu=1,NDIM-1
  do nu=mu+1,NDIM
    ipl=imunu(mu,nu)

    call       assign_shift(w1e,    uo%mu(nu), mu)    ! u_nu(n+mu)
    call       assign_shift(w2e,    uo%mu(mu), nu)    ! u_mu(n+nu)
    call assign_conjg2_mult(wl34e%munu(ipl),w1e,w2e)  ! u_nu(n+mu) * u_mu(n+nu)^+

    call       assign_shift(w1e,ue%mu(mu),-nu,-mu)    ! u_mu(n-nu-mu)
    call       assign_shift(w2e,ue%mu(nu),-mu,-nu)    ! u_nu(n-mu-nu)
    call assign_conjg1_mult(wl98e%munu(ipl),w1e,w2e)  ! u_mu(n-nu-mu)^+ * u_nu(n-mu-nu)
  enddo
  enddo
  deallocate(w1e,w2e)

  do ipl=1,NDIM*(NDIM-1)/2
    call copy_boundary(wl34e%munu(ipl))
    call copy_boundary(wl98e%munu(ipl))
  enddo

  return
end subroutine

!=======================================================================
! SF boundary related routines
!=======================================================================
!========================================================
!
! SF boundary memo
!
! U_4(t=1), U_4(t=2), ..., U_4(t=NTT) : physical
! U_4(t=0)     = U_4(t=NTT) : ghost boundary (not used)
! U_4(t=NTT+1) = U_4(t=1)   : ghost boundary (not used)
!
! U_k(t=1)     = U0  :  SF boundary at t=0
! U_k(t=NTT+1) = U1  :  SF boundary at t=T
! U_k(t=2), U_k(t=3), ..., U_k(t=NTT) : phsyical
!
!
! do itb=1-ieoxyz,NTH-ieoxyz
! it = 2*itb + ieoxyz
! ieoxyz = 0 => itb = 1..NTH => it = 2,4,6,..,NTT
!               itb = 0   => it = 0
!
! ieoxyz = 1 => itb = 0,..NTH-1 => it = 1,3,5,...,NTT-1
!               itb = NTH => it = NTT+1
!               itb = 0   => it = 1
!
!========================================================

subroutine set_sf_boundary_u(u,conjugate)
!
! Set Schroedinger Functional boundary on gauge link.
!
!      this : sf boundary condition
!         u : link field
! conjugate : option : if true, set conjugate boundary condition
!
! u_k(t=1    ,x,y,z) = U0_k <= constant SU(3)
! u_k(t=NTT+1,x,y,z) = U1_k <= constant SU(3)
!
! data on itt=0 ghost sites are not touched.
!
  implicit none
  type(vfield_gluon_wg),    intent(inout) :: u
  logical, optional,        intent(in) :: conjugate
  logical :: is_conjg
  if (.not.present(conjugate)) then
    is_conjg = .false.
  else
    is_conjg = conjugate
  endif
  call set_sf_boundary(u%eo(0),is_conjg)
  call set_sf_boundary(u%eo(1),is_conjg)
  return
end subroutine

subroutine set_sf_boundary_u_eo(ue,conjugate)
!
! Set Schroedinger Functional boundary on gauge link.
!
!        ue : link field (even/odd sites)
! conjugate : option : if true, set conjugate boundary condition
!
! u_k(t=1    ,x,y,z) = U0_k <= constant SU(3)
! u_k(t=NTT+1,x,y,z) = U1_k <= constant SU(3)
!
! data on itt=0 ghost sites are not touched.
!
  use sf_lattice_class
  implicit none
  type(vfield_gluon_eo_wg), intent(inout) :: ue
  logical, optional,        intent(in) :: conjugate
  logical :: is_conjg
  type(sf_lattice_world) :: sf_lattice
  complex(DP) :: phase0(COL,NDIM-1),phase1(COL,NDIM-1)
  integer :: mu,ieo,ic,jc,ix,iy,iz,ieoxyz

  if (.not.is_initialized(sf_lattice)) call new(sf_lattice)

  if (.not.present(conjugate)) then
    is_conjg = .false.
  else
    is_conjg = conjugate
  endif

  phase0(:,:) = get_phase0(sf_lattice)
  phase1(:,:) = get_phase1(sf_lattice)
  if (is_conjg) then
    phase0(:,:) = conjg(phase0(:,:))
    phase1(:,:) = conjg(phase1(:,:))
  endif

  ieo = ue%ieo

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,mu,ic)
  do ix=0,NX1
  do iy=0,NY1
  do iz=0,NZ1
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    if (1 == ieoxyz) then
      do mu=1,NDIM-1 ! spatial only
        !============
        ! t = 1
        !============
        ue%mu(mu)%s(  0,iz,iy,ix)%u(:,:) = Z0
        do ic=1,COL
          ue%mu(mu)%s(  0,iz,iy,ix)%u(ic,ic) = phase0(ic,mu)
        enddo
        !============
        ! t = NTT+1
        !============
        ue%mu(mu)%s(NTH,iz,iy,ix)%u(:,:) = Z0
        do ic=1,COL
          ue%mu(mu)%s(NTH,iz,iy,ix)%u(ic,ic) = phase1(ic,mu)
        enddo
      enddo
    endif
  enddo
  enddo
  enddo  ! end of do ix,iy,iz

  return
end subroutine

subroutine set_sf_boundary_p(p)
!
! Set Schroedinger Functional boundary on gauge link momentum (spatial).
! (set zero at itt=1 for spatial link momentum)
!
!  p : gauge link momentum
!
  implicit none
  type(vfield_gluon_wog),  intent(inout) :: p
  call set_sf_boundary(p%eo(0))
  call set_sf_boundary(p%eo(1))
  return
end subroutine

subroutine set_sf_boundary_p_eo(pe)
!
! Set Schroedinger Functional boundary on gauge link momentum (spatial).
! (set zero at itt=1 for spatial link momentum)
!
!  pe : gauge link momentum (even/odd sites)
!
  implicit none
  type(vfield_gluon_eo_wog), intent(inout) :: pe
  integer :: ieo,mu,ix,iy,iz,ieoxyz
  ieo = pe%ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,mu)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
!========================================================
! SF boundary memo (wo ghost sites)
! do itb=1-ieoxyz,NTH-ieoxyz
! itb0 = itb + ieoxyz
! ieoxyz = 0 => itb = 1..NTH => it = 2,4,6,..,NTT
!               itb = 0   => it = 0  (outside of physical sites)
!
! ieoxyz = 1 => itb = 0,..NTH-1 => it = 1,3,5,...,NTT-1
!               itb = NTH => it = NTT+1 (outside of physical sites)
!               itb = 0   => it = 1 => itb0 = 1 
!========================================================
    if (1 == ieoxyz) then
      !==========
      ! t = 1
      !==========
      do mu=1,NDIM-1 ! spatial only
        pe%mu(mu)%s(  1,iz,iy,ix)%u(:,:) = Z0
      enddo
    endif
  enddo
  enddo
  enddo  ! end of do ix,iy,iz

  return
end subroutine

subroutine accum_mult_at_sf_boundary_u(u,coef,itt0,itt1)
!
! multiply scalar at time t on SU(3) field
!
! ue(t,x,y,z) = ue(t,x,y,z)*coef for t = itt0 or (itt0 and itt1:optional)
!
  implicit none
  type(sfield_gluon_wg),    intent(inout) :: u
  real(DP),                 intent(in) :: coef
  integer,                  intent(in) :: itt0
  integer, optional,        intent(in) :: itt1
  if (.not.present(itt1)) then
    call accum_mult_at_sf_boundary_u_eo(u%eo(0),coef,itt0)
    call accum_mult_at_sf_boundary_u_eo(u%eo(1),coef,itt0)
  else
    call accum_mult_at_sf_boundary_u_eo(u%eo(0),coef,itt0,itt1)
    call accum_mult_at_sf_boundary_u_eo(u%eo(1),coef,itt0,itt1)
  endif
  return
end subroutine

subroutine accum_mult_at_sf_boundary_u_eo(ue,coef,itt0,itt1)
!
! multiply scalar at time t on SU(3) field
!
! ue(t,x,y,z) = ue(t,x,y,z)*coef for t = itt0 or (itt0 and itt1:optional)
!
  implicit none
  type(sfield_gluon_eo_wg), intent(inout) :: ue
  real(DP),                    intent(in) :: coef
  integer,                     intent(in) :: itt0
  integer, optional,           intent(in) :: itt1
  integer :: ieo
  integer :: ix,iy,iz,ieoxyz,ith0,ith1,jtt0,jtt1

  ieo = ue%ieo
  if (.not.present(itt1)) then
    ith0 = itt0/2
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,jtt0)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      jtt0 = 2*ith0 + ieoxyz
      if (jtt0 == itt0) then
        ue%s(ith0,iz,iy,ix)%u(:,:) = ue%s(ith0,iz,iy,ix)%u(:,:)*coef
      endif
    enddo
    enddo
    enddo
  else
    ith0 = itt0/2
    ith1 = itt1/2
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,jtt0,jtt1)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      jtt0 = 2*ith0 + ieoxyz
      if (jtt0 == itt0) then
        ue%s(ith0,iz,iy,ix)%u(:,:) = ue%s(ith0,iz,iy,ix)%u(:,:)*coef
      endif
      jtt1 = 2*ith1 + ieoxyz
      if (jtt1 == itt1) then
        ue%s(ith1,iz,iy,ix)%u(:,:) = ue%s(ith1,iz,iy,ix)%u(:,:)*coef
      endif
    enddo
    enddo
    enddo
  endif
  return
end subroutine

end module
