#ifndef _GPU_BLOCK_SOLVER_CONFIG_H
#define _GPU_BLOCK_SOLVER_CONFIG_H

#define _GPUVERSION_ "GPUBlockSolverClass_v1.0"
!/////////////////////////////////////////////////////////////////
!//
!// CUDA GPU Solver for even/odd preconditioned clover fermions
!// Work with single GPU.
!// Multiple GPUs can be used in embarassingly parrallel manner.
!//
!// gpu1, gpu2, gpu3 solvers are compiled. 
!// They can be submited to each GPU using OpenMP multi threads.
!//
!// This does not work in MPI parallel.
!//
!/////////////////////////////////////////////////////////////////

!////////////////////////////////////////////////////////////
!// number of GPU cards for blocked solver
!////////////////////////////////////////////////////////////
#define _NUM_GPUS (1)


!////////////////////////////////////////////////////////////
!// even/odd site preconditioned
!////////////////////////////////////////////////////////////
#define _NTH (_NTT/2)

!////////////////////////////////////////////////////////////
!// CUDA block size (number of threads in each block)
!// (_NBX*_NBY*_NBZ*_NBT) = THREADS
!////////////////////////////////////////////////////////////
!//#define _NBX 1
!//#define _NBY 1
!//#define _NBZ (_NTZ/_NDIMZ/2)
!//#define _NBT (_NTT/2)

!// for 6^4
!//#define _NBX 1
!//#define _NBY 6
!//#define _NBZ 6
!//#define _NBT 3

!// for 12^4
!//#define _NBX 1
!//#define _NBY 1
!//#define _NBZ 12
!//#define _NBT 6

!// for 8^4
!//#define _NBX 1
!//#define _NBY 4
!//#define _NBZ 8
!//#define _NBT 4

!// for 8^3x16
#define _NBX 1
#define _NBY 4
#define _NBZ 8
#define _NBT 8

!// for 16^3x32
!//#define _NBX 1
!//#define _NBY 1
!//#define _NBZ 16
!//#define _NBT 16

!// for 16^4
!//#define _NBX 1
!//#define _NBY 1
!//#define _NBZ 16
!//#define _NBT 8

!// for 24^4
!//#define _NBX 1
!//#define _NBY 1
!//#define _NBZ 12
!//#define _NBT 12

!// for 32^4
!//#define _NBX 1
!//#define _NBY 1
!//#define _NBZ 16
!//#define _NBT 16

!////////////////////////////////////////////////////////////
!// CUDA Reduction vector length
!// CEIL should be 2^x and satisfy CEIL/2 < THREADS <= CEIL
!////////////////////////////////////////////////////////////
!//#define CEIL 128
#define CEIL 256

!////////////////////////////////////////////////////////////
!// CUDA grid size (number of blocks in each grid
!////////////////////////////////////////////////////////////
#define _NGX (_NTX/_NBX)
#define _NGY (_NTY/_NBY)
#define _NGZ (_NTZ/_NBZ)
#define _NGT (_NTH/_NBT)

!////////////////////////////////////////////////////////////
!// CUDA thread number: 
!// (_NTX*_NTZ*_NTY*_NTX) % NTHREADS should be 0
!////////////////////////////////////////////////////////////
#define THREADS (_NBX*_NBY*_NBZ*_NBT)
#define  BLOCKS ((_NTX*_NTY*_NTZ*_NTH)/THREADS)

#endif
