#include "config.h"
#include "gpu_config.h"

module gpu_bind_class
  use iso_c_binding
  use comlib
  use lattice_class
  use counter_class
  implicit none
  private

  public ::   bind_cpu_to_gpu
  public :: unbind_cpu_to_gpu
  public :: gpu_mult_count

  integer, parameter :: MAXGPU = _MAXGPU
  logical, save :: m_is_gpu_binded_to_cpu = .false.
  integer, save :: m_num_total_gpus = 0
  integer, save :: m_num_use_gpus = 0

  integer(C_INT), BIND(C,name="cu_mult_count"), save :: gpu_mult_count(MAXGPU) = 0
  integer(C_INT), BIND(C,name="cu_nodeid"),     save :: gpu_nodeid
  integer(C_INT), BIND(C,name="cu_ipeo"),       save :: gpu_ipeo
  integer(C_INT), BIND(C,name="cu_nodeidup"),   save :: gpu_nodeidup(0:NDIM-2)
  integer(C_INT), BIND(C,name="cu_nodeiddn"),   save :: gpu_nodeiddn(0:NDIM-2)

  interface 
   !
   ! Fortran interfaces for CPU-GPU binding code using F2003 C binding
   !

    subroutine bind_cpu_to_gpu_loc(num_total_gpus,num_use_gpus) BIND(C,name="bind_cpu_to_gpu_")
      import C_INT
      implicit none
      integer(C_INT), intent(out) :: num_total_gpus
      integer(C_INT), intent(in)  :: num_use_gpus
    end subroutine

    subroutine unbind_cpu_to_gpu_loc() BIND(C,name="unbind_cpu_to_gpu_")
    end subroutine
  end interface
  
contains

subroutine bind_cpu_to_gpu(num_total_gpus,num_use_gpus)
  implicit none
  integer, intent(out) :: num_total_gpus
  integer, intent(in)  :: num_use_gpus
  integer(C_INT) :: f_num_tot_gpus
  integer(C_INT) :: f_num_use_gpus
  if (m_num_use_gpus /= num_use_gpus) then
    call unbind_cpu_to_gpu()
  endif
  if (.not.m_is_gpu_binded_to_cpu) then
    m_num_use_gpus = num_use_gpus
    f_num_use_gpus = num_use_gpus
    call bind_cpu_to_gpu_loc(f_num_tot_gpus,f_num_use_gpus)
      num_total_gpus = f_num_tot_gpus
    m_num_total_gpus = f_num_tot_gpus
    m_is_gpu_binded_to_cpu = .true.
  endif
  gpu_mult_count(:) = 0
  gpu_nodeid = nodeid
  gpu_ipeo   = ipeo  
  gpu_nodeidup(0:NDIM-2) = nodeidup(1:NDIM-1)
  gpu_nodeiddn(0:NDIM-2) = nodeiddn(1:NDIM-1)
  return
end subroutine

subroutine unbind_cpu_to_gpu()
  implicit none
  if (m_is_gpu_binded_to_cpu) then
    call unbind_cpu_to_gpu_loc()
    m_is_gpu_binded_to_cpu = .false.
  endif
  gpu_mult_count(:) = 0
  gpu_nodeid = nodeid
  gpu_ipeo   = ipeo  
  gpu_nodeidup(0:NDIM-2) = nodeidup(1:NDIM-1)
  gpu_nodeiddn(0:NDIM-2) = nodeiddn(1:NDIM-1)
  return
end subroutine

end module
