#!/bin/sh
#
# $Id: Convert.sh,v 1.4 2011/10/31 09:07:33 ishikawa Exp $
#
# ILDG config file data conversion shell script
#
# This uses lime-1.2.3 to extract various information from ILDG config files.
#

LIMEBIN=/home/ishikawa/work/ILDGTOOLS/ILDG/bin

input=$1
outdir=$2
fname=$3

command=$0
path=`dirname $command`

if [ $# -ne 3 ];then
  echo "Usage: $0 ILDG_config_file output_dir_name output_file_name"
  exit -1;
fi

tmp1=tmp1.$$
tmp2=tmp2.$$

date

echo "##############################################################"
echo "# ILDG config file data contents"
echo "##############################################################"
${LIMEBIN}/lime_contents $input
echo ""

#
# header part (lattice size)
#
echo "##############################################################"
echo "# ILDG config file data Lattice Size"
echo "##############################################################"

${LIMEBIN}/lime_extract_record $input 1 1  
echo ""

#
# LFN part
#
echo "##############################################################"
echo "# ILDG config file dataa LFN data" 
echo "##############################################################"

LFN=`${LIMEBIN}/lime_extract_record $input 2 1 `
echo $LFN
fhead=`basename $LFN`
output=${fhead}.bin
echo ""

#
# binary data part
#
echo "##############################################################"
echo "# Extracting config binary data" 
echo "##############################################################"

${LIMEBIN}/lime_extract_record $input 1 2  > $tmp1
echo " Done."
echo ""

#
# show action parameters
#
echo "##############################################################"
echo "# Extracting Action paramters"
echo "##############################################################"

Beta=`echo $LFN | sed 's/.*B\(.*\)Kud.*/\1/g;s/\(^.\)/\1\./g'`
Kud=`echo $LFN | sed 's/.*Kud\(.*\)Ks.*/\1/g;s/^0/0\./g'`
Ks=`echo $LFN | sed 's/.*Ks\(.*\)C.*/\1/g;s/^0/0\./g'`
Csw=`echo $LFN | sed 's/.*C\(.*\)-.-.*/\1/g;s/\(^.\)/\1\./g'`
cRG0=-0.331                                                           # for RG Iwasaki gauge action
traj=`echo $LFN | sed 's/.*-.-\(.*\)/\1/g'`
echo " Done."
echo ""

echo "##############################################################"
echo "# Making output directory"
echo "#   "${outdir}/${traj}
echo "##############################################################"
mkdir -p ${outdir}/${traj}
echo " Done."
echo ""

echo "##############################################################"
echo "# Convert binary data to parallel fortran used in this codes"
echo "# This also convert data endian"
echo "##############################################################"
${path}/convert_ildg_config $tmp1 $LFN $Beta $cRG0 $Kud $Ks $Csw $traj $outdir $fname
echo " Done."
echo ""

rm -f $tmp1 $tmp2
date
