#!/usr/bin/ruby
#
# Generates 4-D SSOR Kernel for Hermitian Squared Domainwall operator (D'D)
#
# $Id: GenMultSSOR.rb,v 1.2 2010/09/01 04:21:07 ishikawa Exp $
#

############################################################################
def mult_inv_mass_matrix(mmatP,mmatN,yt,dagger)
#
#  yt = K \ yt
#
print "
  do iw=1,NS
    yy(iw)%y(:,1:2) = #{yt.get_name}(:,1:2) - #{yt.get_name}(:,3:4)
    yy(iw)%y(:,3:4) = #{yt.get_name}(:,1:2) + #{yt.get_name}(:,3:4)
  enddo
  do iw=1,NS
    ww%y(:,:) = Z0
    do jw=1,NS"
if (dagger == 0)
print "
      ww%y(:,1:2) = ww%y(:,1:2) + #{mmatP}(iw,jw) * yy(jw)%y(:,1:2)
      ww%y(:,3:4) = ww%y(:,3:4) + #{mmatN}(iw,jw) * yy(jw)%y(:,3:4) "
else
print "
      ww%y(:,1:2) = ww%y(:,1:2) + #{mmatP}(jw,iw) * yy(jw)%y(:,1:2)
      ww%y(:,3:4) = ww%y(:,3:4) + #{mmatN}(jw,iw) * yy(jw)%y(:,3:4) "
end
print "
    enddo
    #{yt.get_name}(:,1:2) = (+ ww%y(:,1:2) + ww%y(:,3:4))*0.5_DP
    #{yt.get_name}(:,3:4) = (- ww%y(:,1:2) + ww%y(:,3:4))*0.5_DP
  enddo
"
end

def mult_inv_mass_matrix2(mmatP,mmatN,st,yt,dagger)
#
#  st = K \ yt
#
print "
  do iw=1,NS
    yy(iw)%y(:,1:2) = #{yt.get_name}(:,1:2) - #{yt.get_name}(:,3:4)
    yy(iw)%y(:,3:4) = #{yt.get_name}(:,1:2) + #{yt.get_name}(:,3:4)
  enddo
  do iw=1,NS
    ww%y(:,:) = Z0
    do jw=1,NS"
if (dagger == 0)
print "
      ww%y(:,1:2) = ww%y(:,1:2) + #{mmatP}(iw,jw) * yy(jw)%y(:,1:2)
      ww%y(:,3:4) = ww%y(:,3:4) + #{mmatN}(iw,jw) * yy(jw)%y(:,3:4) "
else
print "
      ww%y(:,1:2) = ww%y(:,1:2) + #{mmatP}(jw,iw) * yy(jw)%y(:,1:2)
      ww%y(:,3:4) = ww%y(:,3:4) + #{mmatN}(jw,iw) * yy(jw)%y(:,3:4) "
end
print "
    enddo
    #{st.get_name}(:,1:2) = (+ ww%y(:,1:2) + ww%y(:,3:4))*0.5_DP
    #{st.get_name}(:,3:4) = (- ww%y(:,1:2) + ww%y(:,3:4))*0.5_DP
  enddo
"
end
############################################################################

require 'FieldClasses'

state = SiteRestriction.new("","blocknormal")

amu = Array.new(4)
for i in 1..4
   amu[i-1] = i
end

puts "!================================================="

yt = Spinor.new("yt(iw)%y")
st = Spinor.new("st(iw)%y")
y  = Spinor.new("y")
w  = Spinor.new("w")

z  = FieldDWSpinor.new("z")
b  = FieldDWSpinor.new("q")
x  = FieldDWSpinor.new("dq")
vu = FieldGauge.new("vu","vector")

##############
# st = Hop z
# yt = z
##############
puts "  do iw=1,NS"
puts "    #{st.get_name}(:,:) = Z0"
puts "  enddo"

##########################################
#  st = D'z = (K - Hop/2)' z
##########################################
for mu in amu.reverse
  puts  "!---------------------------"
  print "!  mu=",mu,"\n"
  puts  "!---------------------------"
  ##########################################
  #  st(iw) = st(iw) + (1+gamma_mu)U_mu(n) z(n+mu)
  ##########################################
  puts "  do iw=1,NS"
  puts  y.assign_proj_up(+mu,z.field_shift(+mu))
  puts  w.assign_mult_u_half(vu.dir(mu).field,y,+1)
  puts st.accum_proj_low(+mu,w)
  puts "  enddo"

  ##############################################
  #  st(iw) = st(iw) + (1-gamma_mu)U_mu(n-mu)^+ z(n-mu)
  ##############################################
  puts "  do iw=1,NS"
  puts  y.assign_proj_up(-mu,z.field_shift(-mu))
  puts  w.assign_mult_u_half(vu.dir(mu).field_shift(-mu),y,-1)
  puts st.accum_proj_low(-mu,w)
  puts "  enddo"
end

##################################################
#  yt(:) = K'(:,:) z
##################################################
mult_inv_mass_matrix2("this%KP","this%KN",yt,z.field,1)

##################################################
# st(iw) = (yt - st(iw)/2) = D'D x = D'z
##################################################
puts ""
puts "  do iw=1,NS"
puts "    #{st.get_name}(:,:) = #{yt.get_name}(:,:) - 0.5_DP * #{st.get_name}(:,:)"
puts "  enddo"
puts ""

#################################################
# multiply inverse diagnal part of (D'D)
#  st = Diag \ st = Diag \ D' z
#################################################
mult_inv_mass_matrix("this%INV_NORMAL_DIAGP","this%INV_NORMAL_DIAGN",st,0)

#################################################
# overrelaxation
# st = b - st * omega = b - (Diag \ D' z) * omega
#################################################
puts "  do iw=1,NS"
puts "    #{st.get_name}(:,:) = #{b.field.get_name}(:,:) - #{st.get_name}(:,:) * omega"
puts "  enddo"
puts ""

puts "if ((ibx == 1 .or. ibx == NBX) .or. &"
puts "    (iby == 1 .or. iby == NBY) .or. &"
puts "    (ibz == 1 .or. ibz == NBZ)) then"
puts "!$OMP BARRIER"
puts "endif"

##################################################
# yt = K st
##################################################
mult_inv_mass_matrix2("this%KP","this%KN",yt,st,0)

##################################################
# Update solution
# x(n) = st(n)
##################################################
puts "  do iw=1,NS"
puts "    #{x.field.get_name}(:,:) = #{st.get_name}(:,:)"

##################################################
# Update auxiliary vector z
# z = D x (keep this relation)
##################################################
########################################
# z(n) = z(n) + K st(n) = z(n) + yt(n)
# st(n) = -st(n)/2
########################################

puts "    #{z.field.get_name}(:,:) = #{z.field.get_name}(:,:) + #{yt.get_name}(:,:)"
puts "    #{st.get_name}(:,:) = - #{st.get_name}(:,:) * 0.5_DP"
puts "  enddo"

for mu in amu.reverse

  puts  "!---------------------------"
  print "!  mu=",mu,"\n"
  puts  "!---------------------------"
  puts ""
  ###############################################################
  # z(n-mu) = z(n-mu) + (1-gamma_mu)U_mu(n-mu) st(n)
  ###############################################################
  puts "  do iw=1,NS"
  puts y.assign_proj_up(-mu,st)
  puts w.assign_mult_u_half(vu.dir(mu).field_shift(-mu),y,+1)
  puts z.field_shift(-mu).accum_proj_low(-mu,w)
  puts "  enddo"

  ################################################################
  # z(n+mu) = z(n+mu) + (1+gamma_mu)U_mu(n)^dag st(n)
  ################################################################
  puts "  do iw=1,NS"
  puts y.assign_proj_up(+mu,st)
  puts w.assign_mult_u_half(vu.dir(mu).field,y,-1)
  puts z.field_shift(+mu).accum_proj_low(+mu,w)
  puts "  enddo"
end

