#!/usr/bin/ruby
#
# Generates 4-D SSOR Kernel for Hermitian Squared Domainwall operator (D'D)
#
# $Id: GenMultSSOR.rb,v 1.2 2010/09/01 04:21:07 ishikawa Exp $
#

############################################################################
def mult_inv_mass_matrix(mmatL,mmatR,yt,dagger)
y  = Spinor.new("y")
w  = Spinor.new("w")
print "
  do iw=1,NS
    yy(:,:,iw) = Z0
    ww(:,:,iw) = Z0
  enddo
  fac = 0.5_DP
  do iw=1,NS
"
puts y.assign_proj_up(-5,yt)  # y := (1-gamma_5)/2 yt
puts w.assign_proj_up(+5,yt)  # w := (1+gamma_5)/2 yt
print "
    do jw=1,NS
"
if (dagger == 1)
print "
      yy(:,1,jw) = yy(:,1,jw) + #{mmatR}(jw,iw)*y(:,1)
      yy(:,2,jw) = yy(:,2,jw) + #{mmatR}(jw,iw)*y(:,2)
      ww(:,1,jw) = ww(:,1,jw) + #{mmatL}(jw,iw)*w(:,1)
      ww(:,2,jw) = ww(:,2,jw) + #{mmatL}(jw,iw)*w(:,2)
"
else
print "
      yy(:,1,jw) = yy(:,1,jw) + #{mmatL}(jw,iw)*y(:,1)
      yy(:,2,jw) = yy(:,2,jw) + #{mmatL}(jw,iw)*y(:,2)
      ww(:,1,jw) = ww(:,1,jw) + #{mmatR}(jw,iw)*w(:,1)
      ww(:,2,jw) = ww(:,2,jw) + #{mmatR}(jw,iw)*w(:,2)
"
end
print "
    enddo
  enddo
  do iw=1,NS
    y(:,1) = yy(:,1,iw)
    y(:,2) = yy(:,2,iw)
    w(:,1) = ww(:,1,iw)
    w(:,2) = ww(:,2,iw)
"
puts yt.assign_proj_low(-5,y)
puts  yt.accum_proj_low(+5,w)
print "
  enddo
"
end
############################################################################

require 'FieldClasses'

state = SiteRestriction.new("","blocknormal")

amu = Array.new(4)
for i in 1..4
   amu[i-1] = i
end

puts "!================================================="

yt = Spinor.new("yt(iw)%y")
st = Spinor.new("st(iw)%y")
y  = Spinor.new("y")
w  = Spinor.new("w")

z  = FieldDWSpinor.new("z")
b  = FieldDWSpinor.new("q")
x  = FieldDWSpinor.new("dq")
vu = FieldGauge.new("vu","vector")

##############
# st = Hop z
##############
puts "  do iw=1,NS"
puts "    st(iw)%y(:,:) = Z0"
puts "  enddo"

##########################################
#  st = D'z
##########################################
for mu in amu.reverse
  puts  "!---------------------------"
  print "!  mu=",mu,"\n"
  puts  "!---------------------------"
  ##########################################
  #  st(iw) = st(iw) + (1+gamma_mu)U_mu(n) z(n+mu)
  ##########################################
  puts "  do iw=1,NS"
  puts  y.assign_proj_up(+mu,z.field_shift(+mu))
  puts  w.assign_mult_u_half(vu.dir(mu).field,y,+1)
  puts st.accum_proj_low(+mu,w)
  puts "  enddo"

  ##############################################
  #  st(iw) = st(iw) + (1-gamma_mu)U_mu(n-mu)^+ z(n-mu)
  ##############################################
  puts "  do iw=1,NS"
  puts  y.assign_proj_up(-mu,z.field_shift(-mu))
  puts  w.assign_mult_u_half(vu.dir(mu).field_shift(-mu),y,-1)
  puts st.accum_proj_low(-mu,w)
  puts "  enddo"
end

##################################################
#  st(:) = K'(:,:) \ st(:)
##################################################
mult_inv_mass_matrix("m_invKL","m_invKR",st,1)

##################################################
# st(iw) = (z - kappa * st(iw)) = D'D x = D'z
##################################################
puts ""
puts "  do iw=1,NS"
puts "    st(iw)%y(:,:) = " + z.field.get_name + "(:,:) - kappa*st(iw)%y(:,:)"
puts "  enddo"
puts ""

#################################################
# multiply inverse diagnal part of (D'D)
#  st = Diag \ st = Diag \ D' z
#################################################
mult_inv_mass_matrix("m_invXL","m_invXR",st,0)

#################################################
# overrelaxation
# st = b - st*omega = b - Diag \ D' z
#################################################
puts "  do iw=1,NS"
puts "    st(iw)%y(:,:) = " + b.field.get_name + "(:,:) - st(iw)%y(:,:)*omega"
puts "  enddo"
puts ""

##################################################
# Update solution
# x(n) = st(n)
##################################################
puts "if ((ibx == 1 .or. ibx == NBX) .or. &"
puts "    (iby == 1 .or. iby == NBY) .or. &"
puts "    (ibz == 1 .or. ibz == NBZ)) then"
puts "!$OMP BARRIER"
puts "endif"

puts "  do iw=1,NS"
puts "    " + x.field.get_name + "(:,:) = st(iw)%y(:,:)"

##################################################
# Update auxiliary vector z
# z = D x (keep this relation)
##################################################
#######################
# z(n) = z(n) + st(n)
#######################

puts "    " + z.field.get_name + "(:,:) = " + z.field.get_name + "(:,:) + st(iw)%y(:,:)"
puts "  enddo"


##################################################
#  st(:) = K(:,:) \ st(:)
##################################################
mult_inv_mass_matrix("m_invKL","m_invKR",st,0)

puts ""
puts "  do iw=1,NS"
puts "    st(iw)%y(:,:) = -kappa*st(iw)%y(:,:)"
puts "  enddo"


for mu in amu.reverse

  ###############################################################
  # z(n-mu) = z(n-mu) + (1-gamma_mu)U_mu(n-mu) st(n)
  ###############################################################
  puts  "!---------------------------"
  print "!  mu=",mu,"\n"
  puts  "!---------------------------"
  puts ""
  puts "  do iw=1,NS"
  puts y.assign_proj_up(-mu,st)
  puts w.assign_mult_u_half(vu.dir(mu).field_shift(-mu),y,+1)
  puts z.field_shift(-mu).accum_proj_low(-mu,w)
  puts "  enddo"

  ################################################################
  # z(n+mu) = z(n+mu) - kappa * (1+gamma_mu)U_mu(n)^dag st(n)
  ################################################################
  puts "  do iw=1,NS"
  puts y.assign_proj_up(+mu,st)
  puts w.assign_mult_u_half(vu.dir(mu).field,y,-1)
  puts z.field_shift(+mu).accum_proj_low(+mu,w)
  puts "  enddo"
end
