#include <config.h>

module main_singlet_class
  use config_profile_class
  use timer_class
  use comlib
  use filename
  use lattice_class
  use solver_parameter_class
  use inverse_class
  use logging_class
  use gauge_field_class
  use gauge_field_dd_class
  use single_prec_class
  use wilson_quark_field_class
  use wilson_quark_field_dd_class
  use wilson_quark_dd_class
#ifdef _CLOVER
  use s_cchlf_class
  use clover_quark_class
  use clover_quark_dd_class
#endif
  use wilson_quark_parameter_class
  use gauge_dd_class
  use quark_prop_class
  use wavefunc_class
  use main_hadron_class
  implicit none
  private

  public :: main_hadron_obj, main_singlet_obj
  public :: new,read,print,delete

  type main_singlet_obj
    type(main_hadron_obj) :: hadrons
    integer :: random_number_seed=1    ! stochastic trace estimator, noise seed
    integer :: nnoise=1                ! stochastic trace estimator, noise number
    character(len=_CHARLEN_) :: seed_path
    character(len=_CHARLEN_) :: seed_name
    type(timer_obj) :: fft_time
  end type

  interface new
    module procedure new_main_singlet
  end interface

  interface delete
    module procedure delete_main_singlet
  end interface

  interface read
    module procedure read_main_singlet
  end interface

  interface print
    module procedure print_main_singlet
  end interface

contains

subroutine new_main_singlet(this)
  implicit none
  type(main_singlet_obj), intent(inout) :: this

  call new(this%hadrons)
  call new(this%fft_time)
  this%seed_path=""
  this%seed_name="singlet_seed."

  return
end subroutine

subroutine delete_main_singlet(this)
  implicit none
  type(main_singlet_obj), intent(inout) :: this

  call tac(this%fft_time)
  call average(this%fft_time)
  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'(" FFT ETime:",E24.16)')this%fft_time%elapse
  endif
  call delete(this%hadrons)

  return
end subroutine

subroutine read_main_singlet(iout,this)
  implicit none
  integer, intent(in) :: iout
  type(main_singlet_obj), intent(inout) :: this

  call read(iout,this%hadrons)

  if (nodeid == 0) then
    read(iout,*)
    read(iout,*)this%random_number_seed
    read(iout,'(A)')this%seed_path
    read(iout,'(A)')this%seed_name
    read(iout,'(A)')this%nnoise
    close(iout)
  endif

  if (NPU > 1) then
    call comlib_bcast(this%random_number_seed,0)
    call comlib_bcast(this%seed_path,0)
    call comlib_bcast(this%seed_name,0)
    call comlib_bcast(this%nnoise,0)
  endif

  return
end subroutine

subroutine print_main_singlet(this)
  implicit none
  type(main_singlet_obj), intent(inout) :: this

  call print(this%hadrons)

  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'("   Seed Save path :"A)')TRIM(this%seed_path)
    write(*,'("   Seed Save name :"A)')TRIM(this%seed_name)
    write(*,'(80("="))')
  endif
  
  return
end subroutine

end module
