  type snk_wavefunc_obj
  !
  ! quark sink function parameters
  !
    type(wavefunc_param_obj) :: param
    complex(DP), pointer :: wavefunc(:) => NULL() ! wf(zxy) spatial wave function
  end type

  interface new
    module procedure new_snk_wavefunc
  end interface

  interface delete
    module procedure delete_snk_wavefunc
  end interface

  interface get
    module procedure get_snk_wavefunc
  end interface

contains

subroutine new_snk_wavefunc(this)
  implicit none
  type(snk_wavefunc_obj), intent(inout) :: this

  call new(this%param)
  if (associated(this%wavefunc)) deallocate(this%wavefunc)

  return
end subroutine

subroutine delete_snk_wavefunc(this)
  implicit none
  type(snk_wavefunc_obj), intent(inout) :: this
  call delete(this%param)
  if (associated(this%wavefunc)) deallocate(this%wavefunc)
  return
end subroutine

subroutine get_snk_wavefunc(this)
  implicit none
  type(snk_wavefunc_obj), intent(inout) :: this
  type(src_wavefunc_obj), allocatable :: twf
  integer :: itt,isp,it,it0

  allocate(twf)
  call new(twf)
  twf%param = this%param
  twf%itx0 = 1
  twf%ity0 = 1
  twf%itz0 = 1
  twf%itt0 = 1

  call get(twf)
  do it=1,NT
    itt = it
    if (itt == twf%itt0) then
      do isp=1,NSPACE
        this%wavefunc(isp) = twf%wavefunc(1,1,isp,it)
      enddo
    endif
  enddo

  call delete(twf)
  deallocate(twf)
  return
end subroutine

function imom(ipx,ipy,ipz) result(ipp)
  implicit none
  integer :: ipx,ipy,ipz,ipp
  ! periodic boundary condition
  ipp = 1 + mod(ipz-1+NZ,NZ) + mod(ipy-1+NY,NY)*NZ + mod(ipx-1+NX,NX)*NZ*NY
  return
end function
